import numpy as np

from sonusai.mixture.types import AudioT
from sonusai.mixture.types import AudiosT
from sonusai.mixture.types import ClassCount
from sonusai.mixture.types import MRecord
from sonusai.mixture.types import MRecords
from sonusai.mixture.types import MixtureDatabase
from sonusai.mixture.types import MixtureIDs
from sonusai.mixture.types import Truth


def get_class_count_from_mrecord(mixdb: MixtureDatabase,
                                 mrecord: MRecord,
                                 target_audios: AudiosT,
                                 noise_audio: AudioT) -> (ClassCount, Truth):
    """Computes the number of samples for which each truth index is active for a given sample-based truth input."""
    from sonusai.mixture import get_class_weights_threshold
    from sonusai.mixture import generate_truth

    truth_t = generate_truth(mixdb=mixdb,
                             mrecord=mrecord,
                             target_audios=target_audios,
                             noise_audio=noise_audio)

    class_weights_threshold = get_class_weights_threshold(mixdb)

    class_count = [0] * mixdb.num_classes
    num_classes = mixdb.num_classes
    if mixdb.truth_mutex:
        num_classes -= 1
    for cl in range(num_classes):
        class_count[cl] = int(np.sum(truth_t[:, cl] >= class_weights_threshold[cl]))

    return class_count, truth_t


def get_class_count_from_mixids(mixdb: MixtureDatabase, mixids: MixtureIDs = None) -> ClassCount:
    """Sums the class counts for given mixids."""
    from sonusai.mixture import get_mrecords_from_mixids

    mixtures = get_mrecords_from_mixids(mixdb, mixids)
    return get_class_count_from_mixtures(mixdb, mixtures)


def get_class_count_from_mixtures(mixdb: MixtureDatabase, mrecords: MRecords) -> ClassCount:
    """Sums the class counts for given mixtures."""
    from sonusai import SonusAIError

    total_class_count = [0] * mixdb.num_classes
    for mrecord in mrecords:
        for cl in range(mixdb.num_classes):
            total_class_count[cl] += mrecord.class_count[cl]

    if mixdb.truth_mutex:
        # Compute the class count for the 'other' class
        if total_class_count[-1] != 0:
            raise SonusAIError('Error: truth_mutex was set, but the class count for the last count was non-zero.')
        total_class_count[-1] = sum([sub.samples for sub in mrecords]) - sum(total_class_count)

    return total_class_count



from dataclasses import dataclass
from typing import List
from typing import Union

import numpy as np
from pyaaware import ForwardTransform
from pyaaware import InverseTransform

from sonusai.mixture.types import AudioData
from sonusai.mixture.types import AudioF
from sonusai.mixture.types import AudioT
from sonusai.mixture.types import AudiosT
from sonusai.mixture.types import Feature
from sonusai.mixture.types import ListAudiosT
from sonusai.mixture.types import MRecord
from sonusai.mixture.types import MixtureDatabase
from sonusai.mixture.types import Truth


# NOTE: global object is required for run-time performance; using 'partial' is much slower.
@dataclass
class GlobalAudio:
    mixdb: MixtureDatabase = None


mp_audiodb = GlobalAudio()


def _augment_audio(audio: AudioT, target: bool = False) -> List[AudioT]:
    from sonusai.mixture import apply_augmentation

    if target:
        augmentations = mp_audiodb.mixdb.target_augmentations
    else:
        augmentations = mp_audiodb.mixdb.noise_augmentations

    return [apply_augmentation(audio_in=audio, augmentation=augmentation) for augmentation in augmentations]


def get_augmented_noise_audios(mixdb: MixtureDatabase, show_progress: bool = False) -> ListAudiosT:
    """Get a list of lists of augmented noise audio data."""
    from tqdm import tqdm

    from sonusai.utils import p_tqdm_map

    raw_noise_audios = get_raw_audios(mixdb=mixdb, show_progress=show_progress, target=False)

    names = [mixdb.noises[i].name for i in range(len(mixdb.noises))]
    progress = tqdm(total=len(names), desc='Augment noise audio', disable=not show_progress)

    mp_audiodb.mixdb = mixdb

    return p_tqdm_map(_augment_audio, raw_noise_audios, progress=progress)


def get_raw_audios(mixdb: MixtureDatabase, show_progress: bool = False, target: bool = True) -> List[AudioT]:
    """Get a list of raw audio data."""
    from tqdm import tqdm

    from sonusai.utils import p_tqdm_map

    if target:
        names = [mixdb.targets[i].name for i in range(len(mixdb.targets))]
        progress = tqdm(total=len(names), desc='Read target audio', disable=not show_progress)
    else:
        names = [mixdb.noises[i].name for i in range(len(mixdb.noises))]
        progress = tqdm(total=len(names), desc='Read noise audio', disable=not show_progress)

    raw_audios = p_tqdm_map(read_audio, names, progress=progress)

    return raw_audios


def check_audio_files_exist(mixdb: MixtureDatabase) -> None:
    """Walk through all the noise and target audio files in a mixture database ensuring that they exist."""
    from os.path import exists

    from sonusai import SonusAIError
    from sonusai.mixture import tokenized_expandvars

    for file_index in range(len(mixdb.noises)):
        file_name, _ = tokenized_expandvars(mixdb.noises[file_index].name)
        if not exists(file_name):
            raise SonusAIError(f'Could not find {file_name}')

    for file_index in range(len(mixdb.targets)):
        file_name, _ = tokenized_expandvars(mixdb.targets[file_index].name)
        if not exists(file_name):
            raise SonusAIError(f'Could not find {file_name}')


# def read_raw_target_audio(mixdb: MixtureDatabase, show_progress: bool = False) -> AudiosT:
#     """Read in all audio data beforehand to avoid reading it multiple times in a loop."""
#     from tqdm import tqdm
#
#     from sonusai.utils import p_tqdm_map
#
#     names = [target.name for target in mixdb.targets]
#     progress = tqdm(total=len(names), desc='Read target audio', disable=not show_progress)
#     raw_target_audio = p_tqdm_map(read_audio, names, progress=progress)
#     progress.close()
#
#     return raw_target_audio


def get_target_noise_audio(mixdb: MixtureDatabase,
                           mrecord: MRecord,
                           raw_target_audios: AudiosT = None,
                           augmented_noise_audios: ListAudiosT = None) -> (AudiosT, AudioT):
    """Apply augmentations and return augmented target and noise data."""
    from pyaaware import FeatureGenerator

    from sonusai import SonusAIError
    from sonusai.mixture import apply_augmentation
    from sonusai.mixture import apply_snr_gain
    from sonusai.mixture import pad_to_samples

    if raw_target_audios is None:
        raw_target_audios = get_raw_audios(mixdb)
    if augmented_noise_audios is None:
        augmented_noise_audios = get_augmented_noise_audios(mixdb)

    target_file_index = mrecord.target_file_index
    target_augmentation_index = mrecord.target_augmentation_index
    if len(target_file_index) != len(target_augmentation_index):
        raise SonusAIError('target_file_index and target_augmentation_index are not the same length')

    fg = FeatureGenerator(feature_mode=mixdb.feature)
    if mrecord.samples % fg.ftransform_R != 0:
        raise SonusAIError(f'Number of samples in mixture is not a multiple of {fg.ftransform_R}')

    targets = []
    for idx in range(len(target_file_index)):
        target_augmentation = mixdb.target_augmentations[target_augmentation_index[idx]]

        target = apply_augmentation(audio_in=raw_target_audios[target_file_index[idx]],
                                    augmentation=target_augmentation,
                                    length_common_denominator=mixdb.feature_step_samples)

        target = pad_to_samples(audio_in=target, samples=mrecord.samples)
        targets.append(target)

    noise = augmented_noise_audios[mrecord.noise_file_index][mrecord.noise_augmentation_index]
    noise, _ = get_next_noise(offset_in=mrecord.noise_offset,
                              length=mrecord.samples,
                              audio_in=noise)

    return apply_snr_gain(mrecord=mrecord, targets=targets, noise=noise)


def get_audio_data(mixdb: MixtureDatabase,
                   mrecord: MRecord,
                   raw_target_audios: AudiosT = None,
                   augmented_noise_audios: ListAudiosT = None,
                   compute_truth: bool = True,
                   compute_segsnr: bool = False,
                   frame_based_segsnr: bool = False) -> AudioData:
    from sonusai.mixture import generate_segsnr
    from sonusai.mixture import generate_truth

    targets, noise = get_target_noise_audio(mixdb=mixdb,
                                            mrecord=mrecord,
                                            raw_target_audios=raw_target_audios,
                                            augmented_noise_audios=augmented_noise_audios)

    truth_t = generate_truth(mixdb=mixdb,
                             mrecord=mrecord,
                             target_audios=targets,
                             noise_audio=noise,
                             compute=compute_truth)

    target = sum(targets)

    segsnr_t = generate_segsnr(mixdb=mixdb,
                               mrecord=mrecord,
                               target_audio=target,
                               noise_audio=noise,
                               compute=compute_segsnr,
                               frame_based=frame_based_segsnr)

    mixture = target + noise
    return AudioData(mixture=mixture,
                     targets=targets,
                     noise=noise,
                     truth_t=truth_t,
                     segsnr_t=segsnr_t)


def concatenate_audio_data(results: List[AudioData]) -> AudioData:
    mixture = np.concatenate([results[i].mixture for i in range(len(results))])

    if any(result.truth_t is None for result in results):
        truth_t = None
    else:
        truth_t = np.concatenate([results[i].truth_t for i in range(len(results))])

    targets = [
        np.concatenate([results[i].targets[j] for i in range(len(results))]) for j in range(len(results[0].targets))]

    noise = np.concatenate([results[i].noise for i in range(len(results))])

    if any(result.segsnr_t is None for result in results):
        segsnr_t = None
    else:
        segsnr_t = np.concatenate([results[i].segsnr_t for i in range(len(results))])

    return AudioData(mixture=mixture,
                     targets=targets,
                     noise=noise,
                     truth_t=truth_t,
                     segsnr_t=segsnr_t)


def get_next_noise(offset_in: int, length: int, audio_in: AudioT) -> (AudioT, int):
    audio_out = np.take(audio_in, range(offset_in, offset_in + length), mode='wrap')
    offset_out = (offset_in + length) % len(audio_in)
    return audio_out, offset_out


def read_audio(name: str) -> AudioT:
    import sox

    from sonusai import SonusAIError
    from sonusai import logger
    from sonusai.mixture import tokenized_expandvars
    from sonusai.mixture import BIT_DEPTH
    from sonusai.mixture import CHANNEL_COUNT
    from sonusai.mixture import SAMPLE_RATE

    expanded_name, _ = tokenized_expandvars(name)

    try:
        # Read in and convert to desired format
        inp = sox.Transformer()
        inp.set_output_format(rate=SAMPLE_RATE, bits=BIT_DEPTH, channels=CHANNEL_COUNT)
        return inp.build_array(input_filepath=expanded_name,
                               sample_rate_in=int(sox.file_info.sample_rate(expanded_name)))

    except Exception as e:
        if name != expanded_name:
            logger.error(f'Error reading {name} (expanded: {expanded_name}): {e}')
        else:
            raise SonusAIError(f'Error reading {name}: {e}')


def get_mixture_data_deprecated(mixdb: MixtureDatabase,
                                mrecord: MRecord) -> (AudioT,
                                                      List[Union[np.ndarray, None]],
                                                      AudioT,
                                                      Feature,
                                                      Truth):
    """Get mixture data assuming nothing has been loaded into memory already."""
    from sonusai.mixture import apply_augmentation
    from sonusai.mixture import generate_truth
    from sonusai.mixture import get_feature_data
    from sonusai.mixture import get_truth_indices_for_target
    from sonusai.mixture import pad_to_samples

    target_audios = []
    target_truth_indices = []
    for idx in range(len(mrecord.target_file_index)):
        target_name = mixdb.targets[mrecord.target_file_index[idx]].name
        target_augmentation = mixdb.target_augmentations[mrecord.target_augmentation_index[idx]]

        target_audio = apply_augmentation(audio_in=read_audio(target_name),
                                          augmentation=target_augmentation,
                                          length_common_denominator=mixdb.feature_step_samples)

        target_audio = np.int16(np.round(np.float32(target_audio) * mrecord.target_snr_gain))
        target_audio = pad_to_samples(audio_in=target_audio, samples=mrecord.samples)
        target_audios.append(target_audio)
        target_truth_indices.append(get_truth_indices_for_target(mixdb.targets[mrecord.target_file_index[idx]]))

    augmented_noise_audios = get_augmented_noise_audios(mixdb=mixdb, show_progress=False)
    raw_noise_audio = augmented_noise_audios[mrecord.noise_file_index][mrecord.noise_augmentation_index]
    noise_audio, _ = get_next_noise(offset_in=mrecord.noise_offset,
                                    length=mrecord.samples,
                                    audio_in=raw_noise_audio)

    noise_audio = np.int16(np.round(np.float32(noise_audio) * mrecord.noise_snr_gain))

    truth_t = generate_truth(mixdb=mixdb,
                             mrecord=mrecord,
                             target_audios=target_audios,
                             noise_audio=noise_audio,
                             compute=True)

    mixture_audio = sum(target_audios) + noise_audio

    # Transform target_audios into a list num_classes long such that each entry is the target data per class
    class_audio = []
    for n in range(mixdb.num_classes):
        class_audio.append(None)
        for idx in range(len(target_audios)):
            if n + 1 in target_truth_indices[idx]:
                if class_audio[n] is None:
                    class_audio[n] = target_audios[idx]
                else:
                    class_audio[n] += target_audios[idx]

    feature, truth_f, _ = get_feature_data(mixdb=mixdb,
                                           mrecord=mrecord,
                                           mixture=mixture_audio,
                                           truth_t=truth_t)

    return mixture_audio, class_audio, noise_audio, feature, truth_f


def calculate_transform_from_audio(audio: AudioT, transform: ForwardTransform) -> AudioF:
    """
    Apply forward transform to input audio data [samples] to generate transform data [frames, bins].
    """
    return transform.execute_all(audio).transpose()


def calculate_audio_from_transform(data: AudioF, transform: InverseTransform, trim: bool = True) -> AudioT:
    """
    Apply inverse transform to input transform data [frames, bins] to generate audio data [samples].
    Trim mode removes starting samples so output waveform will be time-aligned with
    input waveform to the transform.
    """
    audio = transform.execute_all(data.transpose())
    if trim:
        audio = audio[transform.N - transform.R:]

    return audio


def add_audio_data_to_h5(file: str, data: AudioData, save_truth: bool = False, save_segsnr: bool = False) -> None:
    import h5py

    with h5py.File(file, 'a') as f:
        datasets = ['mixture', 'targets', 'noise']
        for dataset in datasets:
            if dataset in f:
                del f[dataset]

        f.create_dataset(name='mixture', data=data.mixture, dtype=np.int16)
        f.create_dataset(name='targets', data=data.targets, dtype=np.int16)
        f.create_dataset(name='noise', data=data.noise, dtype=np.int16)

        if save_truth:
            if 'truth_t' in f:
                del f['truth_t']
            f.create_dataset(name='truth_t', data=data.truth_t, dtype=np.float32)

        if save_segsnr:
            if 'segsnr_t' in f:
                del f['segsnr_t']
            f.create_dataset(name='segsnr_t', data=data.segsnr_t, dtype=np.float32)



from dataclasses import dataclass
from typing import List

import numpy as np

from sonusai.mixture.types import AudioT
from sonusai.mixture.types import AudioT
from sonusai.mixture.types import AudiosT
from sonusai.mixture.types import Feature
from sonusai.mixture.types import FeatureData
from sonusai.mixture.types import ListAudiosT
from sonusai.mixture.types import Location
from sonusai.mixture.types import MRecord
from sonusai.mixture.types import MixtureDatabase
from sonusai.mixture.types import MixtureIDs
from sonusai.mixture.types import Segsnr
from sonusai.mixture.types import Truth


# NOTE: global object is required for run-time performance; using 'partial' is much slower.
@dataclass
class GlobalFeature:
    mixdb: MixtureDatabase = None
    location: str = None


mp_feature = GlobalFeature()


def get_feature_data(mixdb: MixtureDatabase,
                     mrecord: MRecord,
                     mixture: AudioT = None,
                     truth_t: Truth = None,
                     segsnr_t: Segsnr = None,
                     raw_target_audios: AudiosT = None,
                     augmented_noise_audios: ListAudiosT = None,
                     compute_truth: bool = True,
                     compute_segsnr: bool = False) -> FeatureData:
    from pyaaware import FeatureGenerator
    from pyaaware import ForwardTransform

    from sonusai import SonusAIError
    from sonusai.mixture import get_audio_data
    from sonusai.mixture import get_feature_frames_in_mrecord
    from sonusai.mixture import get_transform_frames_in_mrecord
    from sonusai.mixture import segsnr_sample_to_frame
    from sonusai.mixture import truth_reduction
    from sonusai.utils import int16_to_float

    segsnr = None
    if mixture is None or (compute_truth and truth_t is None) or (compute_segsnr and segsnr_t is None):
        audio_data = get_audio_data(mixdb=mixdb,
                                    mrecord=mrecord,
                                    raw_target_audios=raw_target_audios,
                                    augmented_noise_audios=augmented_noise_audios,
                                    compute_truth=compute_truth,
                                    compute_segsnr=compute_segsnr,
                                    frame_based_segsnr=True)
        mixture = audio_data.mixture
        truth_t = audio_data.truth_t
        segsnr = audio_data.segsnr_t
    elif compute_segsnr and segsnr_t is not None:
        segsnr = segsnr_sample_to_frame(mixdb, segsnr_t)

    if len(mixture) != mrecord.samples:
        raise SonusAIError(f'Wrong number of samples in mixture')

    fg = FeatureGenerator(feature_mode=mixdb.feature,
                          num_classes=mixdb.num_classes,
                          truth_mutex=mixdb.truth_mutex)

    fft = ForwardTransform(N=fg.ftransform_N, R=fg.ftransform_R, ttype=fg.ftransform_ttype)

    transform_frames = get_transform_frames_in_mrecord(mixdb, mrecord)
    feature_frames = get_feature_frames_in_mrecord(mixdb, mrecord)

    if truth_t is None:
        truth_t = np.zeros((mrecord.samples, mixdb.num_classes), dtype=np.float32)

    feature = np.empty((feature_frames, fg.stride, fg.num_bands), dtype=np.float32)
    truth_f = np.empty((feature_frames, mixdb.num_classes), dtype=np.complex64)

    feature_frame = 0
    for transform_frame in range(transform_frames):
        indices = slice(transform_frame * fg.ftransform_R,
                        (transform_frame + 1) * fg.ftransform_R)
        fd = fft.execute(int16_to_float(mixture[indices]))

        fg.execute(fd, truth_reduction(truth_t[indices], mixdb.truth_reduction_function))

        if fg.eof():
            feature[feature_frame] = fg.feature()
            truth_f[feature_frame] = fg.truth()
            feature_frame += 1

    if np.isreal(truth_f).all():
        truth_f = np.float32(np.real(truth_f))

    return FeatureData(feature=feature,
                       truth_f=truth_f,
                       segsnr=segsnr)


def concatenate_feature_data(results: List[FeatureData]) -> FeatureData:
    feature = np.concatenate([results[i].feature for i in range(len(results))])

    if any(result.truth_f is None for result in results):
        truth_f = None
    else:
        truth_f = np.concatenate([results[i].truth_f for i in range(len(results))])

    if any(result.segsnr is None for result in results):
        segsnr = None
    else:
        segsnr = np.concatenate([results[i].segsnr for i in range(len(results))])

    return FeatureData(feature=feature,
                       truth_f=truth_f,
                       segsnr=segsnr)


def _get_feature_data_from_dir_kernel(mixid: int) -> FeatureData:
    """Get feature/truth for a given mixid from a directory containing genft data"""
    from os.path import join

    import h5py

    with h5py.File(join(mp_feature.location, mp_feature.mixdb.mixtures[mixid].name), 'r') as f:
        feature = np.array(f['feature'])
        truth_f = np.array(f['truth_f'])
        if 'segsnr' in f:
            segsnr = np.array(f['segsnr'])
        else:
            segsnr = None

    return FeatureData(feature=feature,
                       truth_f=truth_f,
                       segsnr=segsnr)


def get_feature_data_from_dir(mixdb: MixtureDatabase,
                              location: Location,
                              mixids: MixtureIDs = None) -> FeatureData:
    """Get feature/truth for given mixids from a directory containing genft data"""
    from sonusai.mixture import convert_mixids_to_list
    from sonusai.utils import p_map

    mp_feature.mixdb = mixdb
    mp_feature.location = location
    mixids = convert_mixids_to_list(mixdb, mixids)
    results = p_map(_get_feature_data_from_dir_kernel, mixids)

    return concatenate_feature_data(results)


@dataclass(frozen=True)
class FeatureStats:
    feature_ms: float
    feature_samples: int
    feature_step_ms: float
    feature_step_samples: int
    num_bands: int
    stride: int
    step: int
    decimation: int


def get_feature_stats(feature: str, num_classes: int) -> FeatureStats:
    from pyaaware import FeatureGenerator

    import sonusai

    fg = FeatureGenerator(feature_mode=feature, num_classes=num_classes)

    transform_frame_ms = float(fg.ftransform_R) / float(sonusai.mixture.SAMPLE_RATE / 1000)

    return FeatureStats(feature_ms=transform_frame_ms * fg.decimation * fg.stride,
                        feature_samples=fg.ftransform_R * fg.decimation * fg.stride,
                        feature_step_ms=transform_frame_ms * fg.decimation * fg.step,
                        feature_step_samples=fg.ftransform_R * fg.decimation * fg.step,
                        num_bands=fg.num_bands,
                        stride=fg.stride,
                        step=fg.step,
                        decimation=fg.decimation)


def get_feature_from_audio(audio: AudioT, feature: str) -> Feature:
    from sonusai.mixture import get_pad_length
    from sonusai.mixture import MRecord

    fs = get_feature_stats(feature=feature, num_classes=1)
    audio = np.pad(array=audio, pad_width=(0, get_pad_length(len(audio), fs.feature_step_samples)))

    mixdb = MixtureDatabase(feature=feature,
                            mixtures=[MRecord(samples=len(audio))],
                            feature_samples=fs.feature_samples,
                            feature_step_samples=fs.feature_step_samples,
                            num_classes=1,
                            truth_mutex=False,
                            truth_reduction_function='max')
    return get_feature_data(mixdb=mixdb,
                            mrecord=mixdb.mixtures[0],
                            mixture=audio,
                            compute_truth=False).feature


def add_feature_data_to_h5(file: str, data: FeatureData, save_segsnr: bool = False) -> None:
    import h5py

    with h5py.File(file, 'a') as f:
        datasets = ['feature', 'truth_f']
        for dataset in datasets:
            if dataset in f:
                del f[dataset]

        f.create_dataset(name='feature', data=data.feature, dtype=np.float32)
        f.create_dataset(name='truth_f', data=data.truth_f, dtype=np.float32)

        if save_segsnr:
            if 'segsnr' in f:
                del f['segsnr']
            f.create_dataset(name='segsnr', data=data.segsnr, dtype=np.float32)



"""sonusai genft

usage: genft [-hvs] (-d MIXDB) [-i MIXID] [-o OUTPUT]

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -d MIXDB, --mixdb MIXDB         Mixture database directory.
    -i MIXID, --mixid MIXID         Mixture ID(s) to generate. [default: *].
    -o OUTPUT, --output OUTPUT      Output directory.
    -s, --segsnr                    Save segsnr. [default: False].

Generate SonusAI feature/truth data from a SonusAI mixture database.

Inputs:
    MIXDB       A SonusAI mixture database directory.
    MIXID       A glob of mixture ID(s) to generate.

Outputs:
    OUTPUT/     A directory containing:
                    <id>.h5:
                        dataset:    feature
                        dataset:    truth_f
                        dataset:    segsnr (optional)
                    <id>.txt
                    genft.log

"""
from dataclasses import dataclass

from sonusai import logger
from sonusai.mixture import AudioT
from sonusai.mixture import AudiosT
from sonusai.mixture import FeatureData
from sonusai.mixture import ListAudiosT
from sonusai.mixture import MixtureDatabase
from sonusai.mixture import MixtureIDs
from sonusai.mixture import Segsnr
from sonusai.mixture import Truth


# NOTE: global object is required for run-time performance; using 'partial' is much slower.
@dataclass
class GlobalGenft:
    mixdb: MixtureDatabase = None
    raw_target_audios: AudiosT = None
    augmented_noise_audios: ListAudiosT = None
    mixture: AudioT = None
    truth_t: Truth = None
    segsnr_t: Segsnr = None
    compute_truth: bool = None
    compute_segsnr: bool = None
    output_dir: str = None


mp_genft = GlobalGenft()


def genft(mixdb: MixtureDatabase,
          mixids: MixtureIDs = None,
          raw_target_audios: AudiosT = None,
          augmented_noise_audios: ListAudiosT = None,
          mixture: AudioT = None,
          truth_t: Truth = None,
          segsnr_t: Segsnr = None,
          compute_truth: bool = True,
          compute_segsnr: bool = False) -> FeatureData:
    import multiprocessing as mp

    from sonusai.mixture import convert_mixids_to_list
    from sonusai.mixture import concatenate_feature_data
    from sonusai.mixture import get_feature_data
    from sonusai.utils import p_map

    mixids = convert_mixids_to_list(mixdb, mixids)
    results = []
    if mp.current_process().daemon:
        for mixid in mixids:
            results.append(get_feature_data(mixdb=mixdb,
                                            mrecord=mixdb.mixtures[mixid],
                                            mixture=mixture,
                                            truth_t=truth_t,
                                            segsnr_t=segsnr_t,
                                            raw_target_audios=raw_target_audios,
                                            augmented_noise_audios=augmented_noise_audios,
                                            compute_truth=compute_truth,
                                            compute_segsnr=compute_segsnr))
    else:
        mp_genft.mixdb = mixdb
        mp_genft.mixture = mixture
        mp_genft.truth_t = truth_t
        mp_genft.segsnr_t = segsnr_t
        mp_genft.raw_target_audios = raw_target_audios
        mp_genft.augmented_noise_audios = augmented_noise_audios
        mp_genft.compute_truth = compute_truth
        mp_genft.compute_segsnr = compute_segsnr

        results = p_map(_genft_kernel, mixids)

    return concatenate_feature_data(results)


def _genft_kernel(mixid: int) -> FeatureData:
    from sonusai.mixture import get_feature_data

    return get_feature_data(mixdb=mp_genft.mixdb,
                            mrecord=mp_genft.mixdb.mixtures[mixid],
                            raw_target_audios=mp_genft.raw_target_audios,
                            augmented_noise_audios=mp_genft.augmented_noise_audios,
                            mixture=mp_genft.mixture,
                            truth_t=mp_genft.truth_t,
                            segsnr_t=mp_genft.segsnr_t,
                            compute_truth=mp_genft.compute_truth,
                            compute_segsnr=mp_genft.compute_segsnr)


def _process_mixture(mixid: int) -> None:
    from os.path import join
    from os.path import splitext

    import h5py
    import numpy as np

    from sonusai.mixture import add_feature_data_to_h5
    from sonusai.mixture import get_mixture_metadata

    output_name = join(mp_genft.output_dir, mp_genft.mixdb.mixtures[mixid].name)
    with h5py.File(output_name, 'a') as f:
        mixture = None
        if 'mixture' in f:
            mixture = np.array(f['mixture'])

        truth_t = None
        if 'truth_t' in f:
            truth_t = np.array(f['truth_t'])

        segsnr_t = None
        if mp_genft.compute_segsnr and 'segsnr_t' in f:
            segsnr_t = np.array(f['segsnr_t'])

    data = genft(mixdb=mp_genft.mixdb,
                 mixids=mixid,
                 raw_target_audios=mp_genft.raw_target_audios,
                 augmented_noise_audios=mp_genft.augmented_noise_audios,
                 mixture=mixture,
                 truth_t=truth_t,
                 segsnr_t=segsnr_t,
                 compute_segsnr=mp_genft.compute_segsnr)

    add_feature_data_to_h5(file=output_name,
                           data=data,
                           save_segsnr=mp_genft.compute_segsnr)

    with open(file=splitext(output_name)[0] + '.txt', mode='w') as f:
        f.write(get_mixture_metadata(mp_genft.mixdb, mixid))


def main():
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    import time
    from os import makedirs
    from os import remove
    from os.path import exists
    from os.path import isdir
    from os.path import join

    from pyaaware import FeatureGenerator
    from tqdm import tqdm

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.mixture import get_augmented_noise_audios
    from sonusai.mixture import get_raw_audios
    from sonusai.mixture import check_audio_files_exist
    from sonusai.mixture import convert_mixids_to_list
    from sonusai.mixture import load_mixdb
    from sonusai.utils import p_tqdm_map
    from sonusai.utils import human_readable_size
    from sonusai.utils import seconds_to_hms

    verbose = args['--verbose']
    mixdb_loc = args['--mixdb']
    mixid = args['--mixid']
    output_dir = args['--output']
    compute_segsnr = args['--segsnr']

    start_time = time.monotonic()

    if output_dir is None:
        output_dir = mixdb_loc

    if exists(output_dir) and not isdir(output_dir):
        remove(output_dir)

    makedirs(output_dir, exist_ok=True)

    create_file_handler(join(output_dir, 'genft.log'))
    update_console_handler(verbose)
    initial_log_messages('genft')

    logger.info(f'\nLoad mixture database from {mixdb_loc}')
    mixdb = load_mixdb(mixdb_loc)
    mixid = convert_mixids_to_list(mixdb, mixid)

    fg = FeatureGenerator(feature_mode=mixdb.feature,
                          num_classes=mixdb.num_classes,
                          truth_mutex=mixdb.truth_mutex)

    total_samples = sum([sub.samples for sub in [mixdb.mixtures[m] for m in mixid]])
    duration = total_samples / sonusai.mixture.SAMPLE_RATE
    total_transform_frames = total_samples // fg.ftransform_R
    total_feature_frames = total_samples // mixdb.feature_step_samples

    logger.info('')
    logger.info(f'Found {len(mixid):,} mixtures to process')
    logger.info(f'{total_samples:,} samples, '
                f'{total_transform_frames:,} transform frames, '
                f'{total_feature_frames:,} feature frames')

    check_audio_files_exist(mixdb)

    mp_genft.mixdb = mixdb
    mp_genft.output_dir = output_dir
    mp_genft.raw_target_audios = get_raw_audios(mixdb)
    mp_genft.augmented_noise_audios = get_augmented_noise_audios(mixdb)
    mp_genft.compute_segsnr = compute_segsnr

    progress = tqdm(total=len(mixid), desc='genft')
    p_tqdm_map(_process_mixture, mixid, progress=progress)
    progress.close()

    logger.info(f'Wrote {len(mixid)} mixtures to {output_dir}')
    logger.info('')
    logger.info(f'Duration: {seconds_to_hms(seconds=duration)}')
    logger.info(f'feature:  {human_readable_size(total_feature_frames * fg.stride * fg.num_bands * 4, 1)}')
    logger.info(f'truth_f:  {human_readable_size(total_feature_frames * mixdb.num_classes * 4, 1)}')
    if compute_segsnr:
        logger.info(f'segsnr:   {human_readable_size(total_transform_frames * 4, 1)}')

    end_time = time.monotonic()
    logger.info(f'Completed in {seconds_to_hms(seconds=end_time - start_time)}')


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



"""sonusai genmix

usage: genmix [-hvts] (-d MIXDB) [-i MIXID] [-o OUTPUT]

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -d MIXDB, --mixdb MIXDB         Mixture database directory.
    -i MIXID, --mixid MIXID         Mixture ID(s) to generate. [default: *].
    -o OUTPUT, --output OUTPUT      Output directory.
    -t, --truth                     Save truth_t. [default: False].
    -s, --segsnr                    Save segsnr_t. [default: False].

Generate SonusAI mixture data from a SonusAI mixture database.

Inputs:
    MIXDB       A SonusAI mixture database directory.
    MIXID       A glob of mixture ID(s) to generate.

Outputs:
    OUTPUT/     A directory containing:
                    <id>.h5:
                        dataset:    mixture
                        dataset:    target
                        dataset:    noise
                        dataset:    truth_t
                        dataset:    segsnr_t
                    <id>.txt
                    genmix.log
"""
from dataclasses import dataclass

from sonusai import logger
from sonusai.mixture import AudioData
from sonusai.mixture import AudiosT
from sonusai.mixture import ListAudiosT
from sonusai.mixture import MixtureDatabase
from sonusai.mixture import MixtureIDs


# NOTE: global object is required for run-time performance; using 'partial' is much slower.
@dataclass
class GlobalGenmix:
    mixdb: MixtureDatabase = None
    raw_target_audios: AudiosT = None
    augmented_noise_audios: ListAudiosT = None
    compute_truth: bool = None
    compute_segsnr: bool = None
    output_dir: str = None


mp_genmix = GlobalGenmix()


def genmix(mixdb: MixtureDatabase,
           mixids: MixtureIDs = None,
           raw_target_audios: AudiosT = None,
           augmented_noise_audios: ListAudiosT = None,
           compute_truth: bool = False,
           compute_segsnr: bool = False) -> AudioData:
    import multiprocessing as mp

    from sonusai.mixture import concatenate_audio_data
    from sonusai.mixture import convert_mixids_to_list
    from sonusai.mixture import get_audio_data
    from sonusai.utils import p_map

    mixids = convert_mixids_to_list(mixdb, mixids)
    results = []
    if mp.current_process().daemon:
        for mixid in mixids:
            results.append(get_audio_data(mixdb=mixdb,
                                          mrecord=mixdb.mixtures[mixid],
                                          raw_target_audios=raw_target_audios,
                                          augmented_noise_audios=augmented_noise_audios,
                                          compute_truth=compute_truth,
                                          compute_segsnr=compute_segsnr))
    else:
        mp_genmix.mixdb = mixdb
        mp_genmix.raw_target_audios = raw_target_audios
        mp_genmix.augmented_noise_audios = augmented_noise_audios
        mp_genmix.compute_truth = compute_truth
        mp_genmix.compute_segsnr = compute_segsnr

        results = p_map(_genmix_kernel, mixids)

    return concatenate_audio_data(results)


def _genmix_kernel(mixid: int) -> AudioData:
    from sonusai.mixture import get_audio_data

    return get_audio_data(mixdb=mp_genmix.mixdb,
                          mrecord=mp_genmix.mixdb.mixtures[mixid],
                          raw_target_audios=mp_genmix.raw_target_audios,
                          augmented_noise_audios=mp_genmix.augmented_noise_audios,
                          compute_truth=mp_genmix.compute_truth,
                          compute_segsnr=mp_genmix.compute_segsnr)


def _process_mixture(mixid: int) -> None:
    from os.path import join
    from os.path import splitext

    from sonusai import SonusAIError
    from sonusai.mixture import add_audio_data_to_h5
    from sonusai.mixture import get_mixture_metadata

    data = genmix(mixdb=mp_genmix.mixdb,
                  mixids=mixid,
                  raw_target_audios=mp_genmix.raw_target_audios,
                  augmented_noise_audios=mp_genmix.augmented_noise_audios,
                  compute_segsnr=mp_genmix.compute_segsnr,
                  compute_truth=mp_genmix.compute_truth)

    samples = data.mixture.shape[0]
    if mp_genmix.compute_truth:
        if samples != data.truth_t.shape[0]:
            raise SonusAIError(
                f'truth_t samples does not match mixture samples: {data.truth_t.shape[0]} != {samples}')
        if mp_genmix.mixdb.num_classes != data.truth_t.shape[1]:
            raise SonusAIError(
                f'truth_t num_classes is incorrect: {data.truth_t.shape[1]} != {mp_genmix.mixdb.num_classes}')
    for target in data.targets:
        if samples != target.shape[0]:
            raise SonusAIError(f'target samples does not match mixture samples: {target.shape[0]} != {samples}')
    if samples != data.noise.shape[0]:
        raise SonusAIError(f'noise samples does not match mixture samples: {data.noise.shape[0]} != {samples}')
    if mp_genmix.compute_segsnr and samples != data.segsnr_t.shape[0]:
        raise SonusAIError(f'segsnr_t samples does not match mixture samples: {data.segsnr_t.shape[0]} != {samples}')

    output_name = join(mp_genmix.output_dir, mp_genmix.mixdb.mixtures[mixid].name)
    add_audio_data_to_h5(file=output_name,
                         data=data,
                         save_truth=mp_genmix.compute_truth,
                         save_segsnr=mp_genmix.compute_segsnr)

    with open(file=splitext(output_name)[0] + '.txt', mode='w') as f:
        f.write(get_mixture_metadata(mp_genmix.mixdb, mixid))


def main():
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    import time
    from os import makedirs
    from os import remove
    from os.path import exists
    from os.path import isdir
    from os.path import join
    from os.path import splitext

    from tqdm import tqdm

    import sonusai
    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import logger
    from sonusai import update_console_handler
    from sonusai.mixture import check_audio_files_exist
    from sonusai.mixture import convert_mixids_to_list
    from sonusai.mixture import get_augmented_noise_audios
    from sonusai.mixture import get_raw_audios
    from sonusai.mixture import load_mixdb
    from sonusai.utils import human_readable_size
    from sonusai.utils import p_tqdm_map
    from sonusai.utils import seconds_to_hms

    verbose = args['--verbose']
    mixdb_loc = args['--mixdb']
    mixid = args['--mixid']
    output_dir = args['--output']
    compute_segsnr = args['--segsnr']
    compute_truth = args['--truth']

    if not output_dir:
        output_dir = splitext(mixdb_loc)[0]

    if exists(output_dir) and not isdir(output_dir):
        remove(output_dir)

    makedirs(output_dir, exist_ok=True)

    start_time = time.monotonic()

    create_file_handler(join(output_dir, 'genmix.log'))
    update_console_handler(verbose)
    initial_log_messages('genmix')

    logger.info(f'\nLoad mixture database from {mixdb_loc}')
    mixdb = load_mixdb(mixdb_loc)
    mixid = convert_mixids_to_list(mixdb, mixid)

    total_samples = sum([sub.samples for sub in [mixdb.mixtures[m] for m in mixid]])
    duration = total_samples / sonusai.mixture.SAMPLE_RATE

    logger.info('')
    logger.info(f'Found {len(mixid):,} mixtures to process')
    logger.info(f'{total_samples:,} samples')

    check_audio_files_exist(mixdb)

    mp_genmix.mixdb = mixdb
    mp_genmix.raw_target_audios = get_raw_audios(mixdb)
    mp_genmix.augmented_noise_audios = get_augmented_noise_audios(mixdb)
    mp_genmix.compute_truth = compute_truth
    mp_genmix.compute_segsnr = compute_segsnr
    mp_genmix.output_dir = output_dir

    progress = tqdm(total=len(mixid), desc='genmix')
    p_tqdm_map(_process_mixture, mixid, progress=progress)
    progress.close()

    logger.info(f'Wrote {len(mixid)} mixtures to {output_dir}')
    logger.info('')
    logger.info(f'Duration: {seconds_to_hms(seconds=duration)}')
    logger.info(f'mixture:  {human_readable_size(total_samples * 2, 1)}')
    if compute_truth:
        logger.info(f'truth_t:  {human_readable_size(total_samples * mixdb.num_classes * 4, 1)}')
    logger.info(f'target:   {human_readable_size(total_samples * 2, 1)}')
    logger.info(f'noise:    {human_readable_size(total_samples * 2, 1)}')
    if compute_segsnr:
        logger.info(f'segsnr:   {human_readable_size(total_samples * 4, 1)}')

    end_time = time.monotonic()
    logger.info(f'Completed in {seconds_to_hms(seconds=end_time - start_time)}')


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



"""sonusai speech_enh

usage: speech_enh [-hvntxw] (-d INPUT) (-p PREDICT) [-o OUTPUT] [-g MINGAIN]

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -d INPUT, --input INPUT         Input data.
    -p PREDICT, --predict PREDICT   Predict directory.
    -o OUTPUT, --output OUTPUT      Output directory.
    -g MINGAIN, --mingain MINGAIN   Minimum target gain threshold. [default: 0.01]
    -n, --noise                     Extract noise in addition to speech and include metrics.
    -t, --truth                     Calculate extraction using truth and include metrics.
    -x, --nowav                     Do not write WAV files.
    -w, --plot                      Write plots PDF for each mixture.

Run speech enhancement using PFILE model or prediction file and dataset MIXDB SonusAI mixture database directory.

Speech is estimated using a given extraction method on the mixture transform, and optionally noise is also
extracted.

TODO: Plot prediction data PFILE and optionally metrics and truth if feature+truth file is provided. PFILE is an HDF5
file with predict dataset or a WAV with a -m MODEL used to run prediction.

Spectrograms are plotted for truth types energy_f, snr_f, and mapped_snr_f.

INPUT is one of the following:

    * WAV
      Using the given model, generate feature data and run prediction. A model file must be
      provided. The MIXID is ignored. If --energy is specified, plot predict data as energy.

    * JSON
      Using the given mixture database, generate feature and truth data. Run prediction if
      a model is given. The MIXID is required. (--energy is ignored.)

    * HDF5
      Using the given mixture database, generate feature and truth data. Run prediction if
      a model is given. The MIXID is required. (--energy is ignored.)

Prediction data will be written to OUTPUT if a model file is given and OUTPUT is specified.

There will be one plot per active truth index. In addition, the top 5 prediction classes are determined and
plotted if needed (i.e., if they were not already included in the truth plots). For plots generated using a
mixture database, then the target will also be displayed. If mixup is active, then each target involved will
be added to the corresponding truth plot.

Inputs:
    MODEL   A SonusAI trained ONNX model file. If a model file is given, prediction data will be
            generated.
    INPUT   A WAV file, or
            a directory containing a SonusAI mixture database

Outputs:
    {INPUT}-plot.pdf or {INPUT}-mix{MIXID}-plot.pdf
    plot.log
    OUTPUT (if MODEL and OUTPUT are both specified)

"""
from typing import List

import matplotlib.pyplot as plt
import numpy as np
from pyaaware import FeatureGenerator

from sonusai import logger
from sonusai.mixture import AudioT
from sonusai.mixture import Feature
from sonusai.mixture import MixtureDatabase
from sonusai.mixture import Predict
from sonusai.mixture import Truth


def abs2(x):
    return x.real ** 2 + x.imag ** 2


def spec_plot(mixture: AudioT,
              feature: Feature,
              predict: Predict = None,
              target: AudioT = None,
              labels: List[str] = None,
              title: str = '') -> plt.figure:
    from sonusai.mixture import SAMPLE_RATE
    from sonusai.utils import int16_to_float

    num_plots = 4 if predict is not None else 2
    fig, ax = plt.subplots(num_plots, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot the waveform
    x_axis = np.arange(len(mixture), dtype=np.float32) / SAMPLE_RATE
    ax[0].plot(x_axis, int16_to_float(mixture), label='Mixture')
    ax[0].set_xlim(x_axis[0], x_axis[-1])
    # ax[0].set_ylim([-1.025, 1.025])
    if target is not None:
        # Plot target time-domain waveform on top of mixture
        color = 'tab:blue'
        ax[0].plot(x - axis, int16_to_float(target), color=color, label='Target')
        ax[0].set_ylabel('magnitude', color=color)
        # ax[0].tick_params(axis='y', labelcolor=color)
    ax[0].set_title('Waveform')

    # Plot the spectrogram
    ax[1].imshow(np.transpose(feature), aspect='auto', interpolation='nearest', origin='lower')
    ax[1].set_title('Feature')

    if predict is not None:
        # Plot and label the model output scores for the top-scoring classes.
        mean_predict = np.mean(predict, axis=0)
        num_classes = predict.shape[-1]
        top_n = min(10, num_classes)
        top_class_indices = np.argsort(mean_predict)[::-1][:top_n]
        ax[2].imshow(np.transpose(predict[:, top_class_indices]), aspect='auto', interpolation='nearest', cmap='gray_r')
        y_ticks = range(0, top_n)
        ax[2].set_yticks(y_ticks, [labels[top_class_indices[x]] for x in y_ticks])
        ax[2].set_ylim(-0.5 + np.array([top_n, 0]))
        ax[2].set_title('Class Scores')

        # Plot the probabilities
        ax[3].plot(predict[:, top_class_indices])
        ax[3].legend(np.array(labels)[top_class_indices], loc='best')
        ax[3].set_title('Class Probabilities')

    fig.suptitle(title)

    return fig


def spec_energy_plot(mixture: AudioT,
                     feature: Feature,
                     target: AudioT = None,
                     truth_f: Truth = None,
                     predict: Predict = None,
                     tp_title: str = '') -> plt.figure:
    from sonusai.mixture import SAMPLE_RATE
    from sonusai.utils import int16_to_float

    num_plots = 2
    if truth_f is not None:
        num_plots += 1
    if predict is not None:
        num_plots += 1

    fig, ax = plt.subplots(num_plots, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot the waveform
    p = 0
    x_axis = np.arange(len(mixture), dtype=np.float32) / SAMPLE_RATE
    ax[p].plot(x_axis, int16_to_float(mixture), label='Mixture', color='mistyrose')
    ax[0].set_ylabel('magnitude', color='tab:blue')
    ax[p].set_xlim(x_axis[0], x_axis[-1])
    # ax[p].set_ylim([-1.025, 1.025])
    if target is not None:
        # Plot target time-domain waveform on top of mixture
        ax[0].plot(x_axis, int16_to_float(target), label='Target', color='tab:blue')
        # ax[0].tick_params(axis='y', labelcolor=color)
    ax[p].set_title('Waveform')

    # Plot the spectrogram
    p += 1
    ax[p].imshow(np.transpose(feature), aspect='auto', interpolation='nearest', origin='lower')
    ax[p].set_title('Feature')

    if truth_f is not None:
        p += 1
        ax[p].imshow(np.transpose(truth_f), aspect='auto', interpolation='nearest', origin='lower')
        ax[p].set_title('Truth ' + tp_title)

    if predict is not None:
        p += 1
        ax[p].imshow(np.transpose(predict), aspect='auto', interpolation='nearest', origin='lower')
        ax[p].set_title('Predict ' + tp_title)

    return fig


def spec3_plot(predict: Predict,
               truth_f: Truth = None,
               feature: Feature = None,
               title: str = '') -> plt.figure:
    """Plot up to three spectrograms:
            * feature (optional)
            * prediction
            * truth (optional)
    """
    num_plots = 1
    if truth_f is not None:
        num_plots += 1
    if feature is not None:
        num_plots += 1

    fig, ax = plt.subplots(num_plots, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot spectrograms
    p = 0
    if feature is not None:
        ax[p].imshow(np.transpose(feature), aspect='auto', interpolation='nearest', origin='lower')
        ax[p].set_title('Feature ' + title)
        p += 1

    if num_plots == 1:
        # plt will not make an array if only 1
        ax.imshow(np.transpose(predict), aspect='auto', interpolation='nearest', origin='lower')
        ax.set_title('Predict ' + title)
    else:
        ax[p].imshow(np.transpose(predict), aspect='auto', interpolation='nearest', origin='lower')
        ax[p].set_title('Predict ' + title)

    if truth_f is not None:
        p += 1
        ax[p].imshow(np.transpose(truth_f), aspect='auto', interpolation='nearest', origin='lower')
        ax[p].set_title('Truth ' + title)

    return fig


def class_plot(mixture: AudioT,
               target: AudioT = None,
               truth_f: Truth = None,
               predict: Predict = None,
               label: str = '') -> plt.figure:
    """Plot mixture waveform with optional prediction and/or truth together in a single plot.
    The target waveform can optionally be provided, and prediction and truth can have multiple classes.

    Inputs:
        mixture     required, [samples, 1]
        target      optional, list of [samples, 1]
        truth_f     optional, [frames, 1]
        predict     optional, [frames, 1]
        label       optional, label name to use when plotting
    """
    from sonusai import SonusAIError
    from sonusai.mixture import SAMPLE_RATE
    from sonusai.utils import int16_to_float

    if mixture.ndim != 1:
        raise SonusAIError('Too many dimensions in mixture')

    if target is not None and target.ndim != 1:
        raise SonusAIError('Too many dimensions in target')

    # Set default to 1 frame when there is no truth or predict data
    frames = 1
    if truth_f is not None and predict is not None:
        if truth_f.ndim != 1:
            raise SonusAIError('Too many dimensions in truth_f')
        t_frames = len(truth_f)

        if predict.ndim != 1:
            raise SonusAIError('Too many dimensions in predict')
        p_frames = len(predict)

        frames = min(t_frames, p_frames)
    elif truth_f is not None:
        if truth_f.ndim != 1:
            raise SonusAIError('Too many dimensions in truth_f')
        frames = len(truth_f)
    elif predict is not None:
        if predict.ndim != 1:
            raise SonusAIError('Too many dimensions in predict')
        frames = len(predict)

    samples = (len(mixture) // frames) * frames

    # x-axis in sec
    x_axis = np.arange(samples, dtype=np.float32) / SAMPLE_RATE

    fig, ax = plt.subplots(1, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot the time-domain waveforms then truth/prediction on second axis
    ax.plot(x_axis, int16_to_float(mixture[0:samples]), color='mistyrose', label='Mixture')
    color = 'red'
    ax.set_xlim(x_axis[0], x_axis[-1])
    # ax.set_ylim([-1.025, 1.025])
    ax.set_ylabel(f'Amplitude', color=color)
    ax.tick_params(axis='y', labelcolor=color)

    # Plot target time-domain waveform
    if target is not None:
        ax.plot(x_axis, int16_to_float(target[0:samples]), color='blue', label='Target')

    # instantiate 2nd y-axis that shares the same x-axis
    if truth_f is not None or predict is not None:
        y_label = 'Truth/Predict'
        if truth_f is None:
            y_label = 'Predict'
        if predict is None:
            y_label = 'Truth'

        ax2 = ax.twinx()

        color = 'black'
        ax2.set_xlim(x_axis[0], x_axis[-1])
        ax2.set_ylim([-0.025, 1.025])
        ax2.set_ylabel(y_label, color=color)
        ax2.tick_params(axis='y', labelcolor=color)

        if truth_f is not None:
            ax2.plot(x_axis, expand_frames_to_samples(truth_f, samples), color='green', label='Truth')

        if predict is not None:
            ax2.plot(x_axis, expand_frames_to_samples(predict, samples), color='brown', label='Predict')

    # set only on last/bottom plot
    ax.set_xlabel('time (s)')

    fig.suptitle(label)

    return fig


def mix3_plot(mixture: AudioT,
              target: AudioT = None,
              wav1: AudioT = None,
              wav2: AudioT = None,
              title: str = '') -> plt.figure:
    """Plot mixture waveform with other waveforms (prediction, truth, etc.) together in a second plot.
    The target waveform can optionally be provided which is overlayed on top of mixture.

    Inputs:
        mixture     required, [samples, 1]
        target      optional, list of [samples, 1]
        wav1        optional, [frames, 1]
        wav2        optional, [frames, 1]
        title       optional, plot title
    """
    from sonusai import SonusAIError
    from sonusai.mixture import SAMPLE_RATE
    from sonusai.utils import int16_to_float

    if mixture.ndim != 1:
        raise SonusAIError('Too many dimensions in mixture')

    if target is not None and target.ndim != 1:
        raise SonusAIError('Too many dimensions in target')

    # Set default to 1 frame when there is no truth or wav2 data
    frames = 1
    if wav1 is not None and wav2 is not None:
        if wav1.ndim != 1:
            raise SonusAIError('Too many dimensions in wav1')
        t_frames = len(wav1)

        if wav2.ndim != 1:
            raise SonusAIError('Too many dimensions in wav2')
        p_frames = len(wav2)

        frames = min(t_frames, p_frames)
    elif wav1 is not None:
        if wav1.ndim != 1:
            raise SonusAIError('Too many dimensions in wav1')
        frames = len(wav1)
    elif wav2 is not None:
        if wav2.ndim != 1:
            raise SonusAIError('Too many dimensions in wav2')
        frames = len(wav2)

    samples = (len(mixture) // frames) * frames

    num_plots = 1
    if wav1 is not None or wav2 is not None:
        num_plots += 1

    # x-axis in sec
    x_axis = np.arange(samples, dtype=np.float32) / SAMPLE_RATE

    fig, ax = plt.subplots(num_plots, 1, constrained_layout=True, figsize=(11, 8.5))

    # Plot the mixture and optional truth waveforms
    p = 0
    ax[p].plot(x_axis, int16_to_float(mixture[0:samples]), color='mistyrose', label='Mixture')
    color = 'red'
    ax[p].set_xlim(x_axis[0], x_axis[-1])
    # ax[p].set_ylim([-1.025, 1.025])
    ax[p].set_ylabel(f'Amplitude', color=color)
    ax[p].tick_params(axis='y', labelcolor=color)

    # Plot target time-domain waveform
    if target is not None:
        ax[p].plot(x_axis, int16_to_float(target[0:samples]), color='blue', label='Target')

    # instantiate 2nd plot
    if wav1 is not None or wav2 is not None:
        p += 1
        y_label = 'wav1/wav2'
        if wav1 is None:
            y_label = 'wav2'
        if wav2 is None:
            y_label = 'wav1'

        if wav1 is not None:
            ax[p].plot(x_axis, expand_frames_to_samples(wav1, samples), color='green', label=y_label)

        if wav2 is not None:
            ax[p].plot(x_axis, expand_frames_to_samples(wav2, samples), color='brown', label=y_label)

    # set only on last/bottom plot
    ax[p].set_xlabel('time (s)')

    fig.suptitle(title)

    return fig


def expand_frames_to_samples(x: np.ndarray, samples: int) -> np.ndarray:
    samples_per_frame = samples // len(x)
    return np.reshape(np.tile(np.expand_dims(x, 1), [1, samples_per_frame]), samples)


def calculate_unmapped_snr_f(mapped_snr_f: np.ndarray,
                             snr_db_mean: np.ndarray,
                             snr_db_std: np.ndarray,
                             limit_en: bool = False,
                             upper_thr: float = .9999,
                             lower_thr: float = .0001) -> np.ndarray:
    """Calculate unmapped SNR energy from mapped SNR per bin/class.
    Input mapped_snr_f typically [frames, bins] (bins = num_classes) i.e., truth_f or predict data
    limiter can be applied as min(upper_thr, mapped_snr_f) and max(lower_thr, mapped_snr_f)
    which is useful for metrics that need to avoid +/-Inf.
    Note: verified as exact match inverse of calculate_map_snr_f() without thresholds
          default thresholds of .0001, .9999 are equiv to snr_f of
    """
    import scipy.special as sc

    old_err = np.seterr(divide='ignore', invalid='ignore')
    if limit_en:
        mapped_snr_f = np.maximum(np.minimum(mapped_snr_f, upper_thr), lower_thr)
    snr_db = np.double(snr_db_std) * np.sqrt(2) * sc.erfinv(2 * np.double(mapped_snr_f) - 1)
    unmapped_snr_f = 10 ** ((snr_db + np.double(snr_db_mean)) / 10)
    np.seterr(**old_err)

    return np.float32(unmapped_snr_f)


def mean_square_error(truth_f: Truth, predict: Predict, squared: bool = False) -> (np.ndarray, np.ndarray, np.ndarray):
    """Calculate mean square error or root-mean-square error between truth_f and predict.
    Typical inputs expected to be size [frames, num_classes|bins]
    Returns: 
        err     mean over both dimensions
        err_c   mean-over-dim0, i.e., a value per class/bin
        err_f   mean-over-dim1, i.e., a value per frame
    """
    sq_err = np.square(truth_f - predict)

    # mean over frames for value per class
    err_c = np.mean(sq_err, axis=0)

    # mean over classes for value per frame
    err_f = np.mean(sq_err, axis=1)

    # mean over all
    err = np.mean(sq_err)

    if not squared:
        err_c = np.sqrt(err_c)
        err_f = np.sqrt(err_f)
        err = np.sqrt(err)

    return err, err_c, err_f


def log_error(truth_f: Truth, predict: Predict) -> (np.ndarray, np.ndarray, np.ndarray):
    """Calculate log error between truth_f and predict
    Typical inputs expected to be size [frames, num_classes|bins]
    Returns:
        err      mean over both dimensions, scalar value
        err_c    mean-over-dim0 i.e. a value per class/bin
        err_f    mean-over-dim1 i.e. a value per frame
    """
    sq_truth_f = np.real(truth_f * np.conjugate(truth_f))
    sq_predict = np.real(predict * np.conjugate(predict))
    log_err = abs(10 * np.log10((sq_truth_f + np.finfo(np.float32).eps) / (sq_predict + np.finfo(np.float32).eps)))

    # mean over frames for value per class
    err_c = np.mean(log_err, axis=0)

    # mean over classes for value per frame
    err_f = np.mean(log_err, axis=1)

    # mean over all
    err = np.mean(log_err)

    return err, err_c, err_f


def calculate_energy_f_from_audio(audio: AudioT, feature_gen: FeatureGenerator) -> np.ndarray:
    """Calculate energy_f from time-domain audio given specified feature generator."""
    from pyaaware import ForwardTransform

    from sonusai.mixture import get_feature_stats
    from sonusai.mixture import get_pad_length
    from sonusai.utils import int16_to_float

    fs = get_feature_stats(feature=feature_gen.feature_mode, num_classes=feature_gen.num_classes)

    audio = np.pad(array=audio, pad_width=(0, get_pad_length(len(audio), fs.feature_step_samples)))

    target_fft = ForwardTransform(N=feature_gen.frame_size * 4, R=feature_gen.frame_size)

    frames = int(len(audio) / fs.feature_step_samples)
    bins = feature_gen.bin_end - feature_gen.bin_start + 1
    energy_f = np.zeros((frames, bins), dtype=np.float32)
    idx = 0
    for offset in range(0, len(audio), feature_gen.frame_size):
        energy_f[idx, :] = np.float32(
            target_fft.energy_f(int16_to_float(audio[offset:offset + feature_gen.frame_size])))
        idx += 1

    return energy_f


def calculate_energy_f_from_mixid(mixdb: MixtureDatabase, mixid: int) -> np.ndarray:
    """Calculate energy_f of mixid mixture waveform in mixdb (energy_f = power spectral density)
    Simply re-generate mixture and transform
    Note: only matches direct bin modes:  b**NN****, i.e. blh28bs1, bcr28bs1, ..."""
    from pyaaware import ForwardTransform

    from sonusai.mixture import get_mixture_data_deprecated
    from sonusai.utils import int16_to_float

    mixture, _, target, noise, _, _ = get_mixture_data_deprecated(mixdb=mixdb, mixid=mixid)

    tf = ForwardTransform(N=mixdb.frame_size * 4, R=mixdb.frame_size)
    mix_energy_f = tf.energy_f(int16_to_float(mixture))
    noise_energy_f = tf.energy_f(int16_to_float(noise))
    target_energy_f = tf.energy_f(int16_to_float(target))

    return mix_energy_f, target_energy_f, noise_energy_f


def mmse_stsa(xi: np.ndarray, gamma: np.ndarray) -> np.ndarray:
    """Computes the MMSE-STSA gain function.
    Arguments:
        xi      a priori SNR
        gamma   a posteriori SNR
    Returns:
        gain    MMSE-STSA gain function
   """
    from scipy.special import i0
    from scipy.special import i1

    old_err = np.seterr(divide='ignore', invalid='ignore', over='ignore')
    nu = np.multiply(xi, np.divide(gamma, np.add(1, xi)))
    # MMSE-STSA gain function
    gain = np.multiply(np.multiply(np.multiply(np.divide(np.sqrt(np.pi), 2),
                                               np.divide(np.sqrt(nu), gamma)), np.exp(np.divide(-nu, 2))),
                       np.add(np.multiply(np.add(1, nu), i0(np.divide(nu, 2))),
                              np.multiply(nu, i1(np.divide(nu, 2)))))
    # replace by Wiener gain
    idx = np.isnan(gain) | np.isinf(gain)
    # Wiener gain
    gain[idx] = np.divide(xi[idx], np.add(1, xi[idx]))
    np.seterr(**old_err)

    return gain


def speech_est_from_snr_f(mixture_tf: np.ndarray,
                          snr_f: np.ndarray,
                          min_gain: np.ndarray,
                          apost_snr: np.ndarray = None,
                          gfunc: callable = mmse_stsa) -> (np.ndarray, np.ndarray, np.ndarray):
    """Compute speech estimate from transform and estimation/prediction data
    Inputs:
        mixture_tf      mixture transform data ([frames, bins] complex float)
        snr_f           a-priori snr_f prediction or truth_f data [frames, bins]
                        frames can be smaller than mixture_tf frames, calculated gain will be zero-padded
        min_gain        minimum gain
        apost_snr       a-posteriori SNR
        gfunc           gain function type, e.g., 'mmse_stsa'
    Outputs:
        est_tf          speech estimate transform data ([frames, bins] complex float)
        gain            gain data [frames, bins], zero-padded in frame dimension if less than mixture_tf frames
        apost_snr       a-posteriori SNR estimated from snr_f (a-priori snr)
    """
    if apost_snr is None:
        # a-posteriori snr estimate, limited to -120 dB
        apost_snr = np.maximum(snr_f + 1, 1e-12)

    gain = np.maximum(gfunc(snr_f, apost_snr), min_gain)
    gain = np.pad(gain, ((0, mixture_tf.shape[0] - gain.shape[0]), (0, 0)))
    est_tf = mixture_tf * gain

    return est_tf, gain, apost_snr


def noise_est_from_snr_f(mixture_tf: np.ndarray,
                         snr_f: np.ndarray,
                         apost_snr: np.ndarray = None) -> (np.ndarray, np.ndarray, np.ndarray):
    """Compute noise estimate from transform and estimation/prediction data (snr_f)
    Inputs:
        mixture_tf      mixture transform data ([frames, bins] complex float)
        snr_f           a-priori snr_f prediction or truth_f data [frames, bins]
                        frames can be smaller than mixture_tf frames, calculated gain will be zero-padded
        apost_snr       a-posteriori SNR estimated from snr_f (a-priori snr)
                        Calculated internally if not provided
    Outputs:
        est_tf          noise estimate transform data ([frames, bins] complex float)
        gain            gain data [frames, bins], zero-padded in frame dimension if less than mixture_tf frames
        apost_snr       a-posteriori SNR estimated from snr_f (a-priori snr)
    """
    if apost_snr is None:
        # a-posteriori snr estimate, limited to -120 dB
        apost_snr = np.maximum(snr_f + 1, 1e-12)

    gain = (1 / np.square(1 + snr_f) + snr_f / ((1 + snr_f) * apost_snr))
    gain = np.pad(gain, ((0, mixture_tf.shape[0] - gain.shape[0]), (0, 0)))
    est_tf = mixture_tf * gain

    return est_tf, gain, apost_snr


def calculate_pesq(speech_ref: np.ndarray, speech_est: np.ndarray, error_value: float = 0) -> float:
    """Computes the PESQ score of speech estimate audio vs. the clean speech estimate audio.
    Upon error, assigns a value of 0, or user specified value in error_value
    Argument/s:
        speech_ref      speech reference audio, np.ndarray vector of samples.
        speech_est      speech estimated audio, np.ndarray vector of samples.
    Returns:
        score           float value between -0.5 and 4.5.
   """
    from pesq import NoUtterancesError
    from pesq import pesq

    try:
        score = pesq(16000, speech_ref, speech_est, mode='wb')
    except NoUtterancesError as e:
        score = error_value

    return score


def run_sonusai_hypermodel(hypermodel, model, feature):
    """Run hypermodel prediction on non-reshaped feature data."""
    from sonusai.data_generator import get_frames_per_batch
    from sonusai.utils import reshape_inputs
    from sonusai.utils import reshape_outputs

    # Pad with zeros in order to create an entire batches of data
    frames = feature.shape[0]
    frames_per_batch = get_frames_per_batch(hypermodel.batch_size, hypermodel.timesteps)
    padding = frames_per_batch - feature.shape[0] % frames_per_batch
    feature = np.pad(array=feature, pad_width=((0, padding), (0, 0), (0, 0)))
    feature, _ = reshape_inputs(feature=feature,
                                batch_size=hypermodel.batch_size,
                                timesteps=hypermodel.timesteps,
                                flatten=hypermodel.flatten,
                                add1ch=hypermodel.add1ch)
    predict = model.predict(feature, verbose=0)
    predict, _ = reshape_outputs(predict=predict, timesteps=hypermodel.timesteps)
    predict = predict[:frames, :]
    return predict


def main():
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    input_name = args['--input']
    predict_dir = args['--predict']
    output_dir = args['--output']
    mingain = float(args['--mingain'])
    noise_estimate_enable = args['--noise']
    truth_estimate_mode = args['--truth']
    disable_wav = args['--nowav']
    plot_en = args['--plot']

    from os import makedirs
    from os.path import basename
    from os.path import isdir
    from os.path import join
    from os.path import splitext

    import h5py
    import keras_tuner as kt
    import pandas as pd
    from matplotlib.backends.backend_pdf import PdfPages
    from pyaaware import ForwardTransform
    from pyaaware import InverseTransform
    from pyaaware import Predict

    from sonusai import SonusAIError
    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import update_console_handler
    from sonusai.mixture import calculate_audio_from_transform
    from sonusai.mixture import calculate_transform_from_audio
    from sonusai.mixture import convert_mixids_to_list
    from sonusai.mixture import get_mixture_data_deprecated
    from sonusai.mixture import get_mrecords_from_mixids
    from sonusai.mixture import load_mixdb
    from sonusai.utils import check_keras_overrides
    from sonusai.utils import create_ts_name
    from sonusai.utils import float_to_int16
    from sonusai.utils import import_keras_model
    from sonusai.utils import int16_to_float
    from sonusai.utils import write_wav

    if output_dir is None:
        output_dir = create_ts_name('speech_enh')

    makedirs(output_dir, exist_ok=True)

    # Setup logging file
    create_file_handler(join(output_dir, 'speech_enh.log'))
    update_console_handler(verbose)
    initial_log_messages('speech_enh')

    if not isdir(predict_dir):
        raise SonusAIError(f'{predict_dir} is not a directory')

    # Setup input file (truth)
    logger.info(f'Input:  {input_name}')
    iext = splitext(input_name)[1]
    ipath = splitext(input_name)[0]
    ibase_name = basename(splitext(input_name)[0])
    if iext == '.json':
        logger.info(f'Input file is .json file, attempting to load mixdb')
        mixdb = load_mixdb(input_name)
        num_mixtures = len(mixdb.mixtures)
        # Review truth settings, use first target assuming all are the same
        settings = mixdb.targets[0].truth_settings[0]
        logger.info(f'Truth function is {settings.function} (est. from first target)')
        # Set up the transforms
        itf = InverseTransform(N=mixdb.itransform.N, R=mixdb.itransform.R, ttype=mixdb.itransform.ttype)
        ftf = ForwardTransform(N=mixdb.ftransform.N, R=mixdb.ftransform.R, ttype=mixdb.ftransform.ttype)

        logger.debug(f'Setup forward transform with N={ftf.N}, R={ftf.R}, ttype={ftf.ttype}')
        logger.debug(f'Setup inverse transform with N={itf.N}, R={itf.R}, ttype={itf.ttype}')
        mixid_padding = mixdb.mixid_padding
        mixid = convert_mixids_to_list(mixdb, mixid)

    else:
        logger.info(f'Truth .h5 input not supported yet.')
        raise SystemExit(1)

    # Setup predict file
    keras_model = None
    onnx_model = None
    pext = splitext(predict_dir)[1]
    pbase_name = basename(splitext(predict_dir)[0])
    if isdir(predict_dir):
        logger.info(f'Prediction is a subdir, assuming it contains a set of prediction .h5 files')
        # TODO
    elif pext == ".wav":
        logger.info(f'Prediction is .wav, attempting model load from:  {model_name}')
        if model_name is not None:
            mbase_name = basename(splitext(model_name)[0])
            try:
                onnx_model = Predict(model_name)
            except Exception as e:
                raise SonusAIError(f'Error loading ONNX model {model_name}, exiting. Msg: {e}')

            logger.debug(f'Model feature name {model.feature}')
            logger.debug(f'Model input shape  {model.input_shape}')
            logger.debug(f'Model output shape {model.output_shape}')
        else:
            raise SonusAIError('Must specify -m .onnx model file when prediction input is .wav')
        output_dir = '.'
    elif pext == ".h5":
        logger.info(f'Prediction file is an HDF5 file, attempting to load predict data')
        with h5py.File(predict_dir, 'r') as f:
            if 'predict' in f:
                predict = np.array(f['predict'])

        if predict.ndim > 2:  # Check if timestep dimension and reshape to remove
            predict = np.reshape(predict, (predict.shape[0] * predict.shape[1], predict.shape[2]))

        # TODO: generate predict_wav here from predict data, assumes transform invertible
        output_dir = '.'
    else:
        # assume it's a SonusAI Keras .py file
        if weights_name is None:
            raise SonusAIError(f'Error: weight file not provided for {predict_dir}, exiting.')
        logger.info('Attempting Sonusai Keras .py hypermodel load')
        try:
            pymodel = import_keras_model(predict_dir)
            hypermodel = pymodel.MyHyperModel()
            if hypermodel.timesteps >= 0:
                timesteps = 1
            else:
                timesteps = 0
            check_keras_overrides(pymodel, mixdb.feature, mixdb.num_classes, timesteps, batch_size)
            hypermodel = pymodel.MyHyperModel(feature=mixdb.feature,
                                              num_classes=mixdb.num_classes,
                                              timesteps=timesteps,
                                              batch_size=batch_size)
            logger.info(f'Building model with batch_size {batch_size} and timesteps {hypermodel.timesteps}')
            keras_model = hypermodel.build_model(kt.HyperParameters())
            logger.info(f'Loading weights from {weights_name}')
            keras_model.load_weights(weights_name)

        except Exception as e:
            logger.exception(f'Error: {predict_dir} hypermodel load or build_model() failed, msg: {e}.')
            raise SystemExit(1)

    basepath = output_dir + '/'

    mrecords = get_mrecords_from_mixids(mixdb, mixid)
    num_mrecords = len(mrecords)

    # Main loop for each mixid run prediction, estimation, metrics
    truth_f = None
    noise_est_f = None
    for mi in range(num_mrecords):
        logger.info(f'Processing mixture {mixid[mi]} ({mi + 1} out of {num_mrecords})')
        # first target index only
        ti = mixdb.mixtures[mi].target_file_index[0]
        # first truth function only
        settings = mixdb.targets[ti].truth_settings[0]
        snr_db_mean = settings.config['snr_db_mean']
        snr_db_std = settings.config['snr_db_std']

        # 1) Generate feature and mixture transform data
        mixture, class_audio, noise, feature, truth_f = get_mixture_data_deprecated(mixdb, mrecords[mi])

        mixture = int16_to_float(mixture)
        mixture_tf = calculate_transform_from_audio(mixture, ftf)

        # TODO: make target, noise, and truth_f optional
        target = int16_to_float(class_audio[0])
        target_tf = calculate_transform_from_audio(target, ftf)

        noise = int16_to_float(noise)
        noise_tf = calculate_transform_from_audio(noise, ftf)

        # 2) Run prediction
        if keras_model is not None:
            predict = run_sonusai_hypermodel(hypermodel, keras_model, feature)
        elif onnx_model is not None:
            # TODO: create function reshape+predict to the onnx model batch, tstep
            predict = model.execute(feature)
        # TODO: add predict data file output

        # 3) Estimate speech and noise tf and .wav
        # Apply default limits for unmapping and metrics
        pred_snr_f = calculate_unmapped_snr_f(predict, snr_db_mean, snr_db_std, True)
        speech_est_f, target_gain, apost_snr = speech_est_from_snr_f(mixture_tf, pred_snr_f, mingain,
                                                                     gfunc='mmse_stsa')
        speech_est_wav = calculate_audio_from_transform(speech_est_f, itf)
        # TODO: make optional if noise enabled cli switch
        noise_est_f, noise_gain, _ = noise_est_from_snr_f(mixture_tf, pred_snr_f, apost_snr)
        noise_est_wav = calculate_audio_from_transform(noise_est_f, itf)

        # 4) Calculate speech and noise metrics (optional, if truth_f avail, and target, noise, etc.)
        if truth_f is not None:
            # mixid string for file outputs
            mixidstr = f'{mi:0{mixid_padding}}'
            # SNR est accuracy: spectral distortion metric effectively the rmse of unmapped snr's in db
            truth_f_db = 10 * np.log10(truth_f + np.finfo(np.float32).eps)
            predict_db = 10 * np.log10(pred_snr_f + np.finfo(np.float32).eps)
            sd_snr, sd_snr_bin, sd_snr_frame = mean_square_error(truth_f_db, predict_db)
            # Speech PSD estimation accuracy symmetric mean log-spectral distortion (logerr):
            lerr_sp, lerr_sp_bin, lerr_sp_frame = log_error(target_tf, speech_est_f)
            # Speech intelligibility measure - PESQ
            pesq_speech = calculate_pesq(target, speech_est_wav)
            pesq_mixture = calculate_pesq(target, mixture)
            # pesq improvement
            pesq_impr = pesq_speech - pesq_mixture
            # pesq improvement %
            pesq_impr_pc = pesq_impr / (pesq_mixture + np.finfo(np.float32).eps) * 100
            if noise_est_f is not None:
                lerr_n, lerr_n_bin, lerr_n_frame = log_error(noise_tf, noise_est_f)

            # 5) Save per mixture metric results
            # Single row in table of scalar metrics per mixture
            mtable1_col = ['MXSNR', 'MXPESQ', 'PESQ', 'PESQi', 'PESQi%', 'SNRSD', 'SPLERR', 'NLERR', 'SPFILE', 'NFILE']
            ni = mixdb.mixtures[mi].noise_file_index
            metr1 = [mixdb.mixtures[mi].snr, pesq_mixture, pesq_speech, pesq_impr, pesq_impr_pc, sd_snr, lerr_sp,
                     lerr_n, basename(mixdb.targets[ti].name), basename(mixdb.noises[ni].name)]
            mtab1 = pd.DataFrame([metr1], columns=mtable1_col, index=[mi])
            # Stats of per frame estimation metrics
            efs_table2_col = ['Max', 'Min', 'Avg', 'Median']
            efs_table2_row = ['SNRSD', 'SPLERR', 'NLERR']
            metr2 = [[np.max(sd_snr_frame), np.min(sd_snr_frame), np.mean(sd_snr_frame), np.median(sd_snr_frame)],
                     [np.max(lerr_sp_frame), np.min(lerr_sp_frame), np.mean(lerr_sp_frame), np.median(lerr_sp_frame)],
                     [np.max(lerr_n_frame), np.min(lerr_n_frame), np.mean(lerr_n_frame), np.median(lerr_n_frame)]]
            mtab2 = pd.DataFrame(metr2, columns=efs_table2_col, index=efs_table2_row)

            mtab2flat_col = ['MXSNR', 'SNRSD Max', 'SNRSD Min', 'SNRSD Avg', 'SNRSD Median',
                             'SPLERR Max', 'SPLERR Min', 'SPLERR Avg', 'SPLERR Median',
                             'NLERR Max', 'NLERR Min', 'NLERR Avg', 'NLERR Median']
            tmp = np.insert(np.array(metr2), 0, mixdb.mixtures[mi].snr).reshape(1, 13)
            mtab2_flat = pd.DataFrame(tmp, columns=mtab2flat_col, index=[mi])
            if 'all_mtab1' in locals():
                # Update all mixture table
                all_mtab1 = pd.concat([all_mtab1, mtab1])
                all_mtab2 = pd.concat([all_mtab2, mtab2_flat])
            else:
                # create the table if it doesn't exist
                all_mtab1 = mtab1
                all_mtab2 = mtab2_flat

            metric_fname = basepath + mixidstr + '-spenh-metrics.txt'
            with open(metric_fname, 'w') as f:
                print('Speech enhancement metrics:', file=f)
                print(mtab1.round(2).to_string(), file=f)
                print('', file=f)
                print(f'Extraction statistics over {mixture_tf.shape[0]} frames:', file=f)
                print(mtab2.round(2).to_string(), file=f)
                print('', file=f)
                print(f'Target path: {mixdb.targets[ti].name}', file=f)
                print(f'Noise path: {mixdb.noises[ni].name}', file=f)
                # print(f'PESQ improvement: {pesq_impr:0.2f}, {pesq_impr_pc:0.1f}%', file=f)

            # 6) Estimate speech from truth to test the extraction-reconstruction methods, and calc metrics
            if truth_estimate_mode is True:
                truth_snr_f = calculate_unmapped_snr_f(truth_f, snr_db_mean, snr_db_std, True)
                speech_trest_f, tru_target_gain, apost_snr = \
                    speech_est_from_snr_f(mixture_tf, truth_snr_f, mingain, gfunc='mmse_stsa')
                speech_trest_wav = calculate_audio_from_transform(speech_trest_f, itf)
                noise_trest_f, noise_gain, _ = noise_est_from_snr_f(mixture_tf, truth_snr_f, apost_snr)
                noise_trest_wav = calculate_audio_from_transform(noise_trest_f, itf)

                # Speech PSD estimation accuracy symmetric mean log-spectral distortion (logerr):
                lerr_sptr, lerr_sptr_bin, lerr_sptr_frame = log_error(target_tf, speech_trest_f)
                # Speech intelligibility measure - PESQ
                pesq_speechtr = calculate_pesq(target, speech_trest_wav)
                # pesq improvement
                pesq_impr_sptr = pesq_speechtr - pesq_mixture
                # pesq improvement %
                pesq_impr_pctr = pesq_impr_sptr / (pesq_mixture + np.finfo(np.float32).eps) * 100
                lerr_ntr, lerr_ntr_bin, lerr_ntr_frame = log_error(noise_tf, noise_trest_f)

                mtable3_col = ['MXSNR', 'MXPESQ', 'PESQ', 'PESQi', 'PESQi%', 'SPLERR', 'NLERR']
                metr3 = [mixdb.mixtures[mi].snr, pesq_mixture, pesq_speechtr, pesq_impr_sptr, pesq_impr_pctr, lerr_sptr,
                         lerr_ntr]
                mtab3 = pd.DataFrame([metr3], columns=mtable3_col, index=[mi])

                # Stats of per frame estimation metrics
                efs_table4_col = ['Max', 'Min', 'Avg', 'Median']
                efs_table4_row = ['SPLERR', 'NLERR']
                metr4 = [[np.max(lerr_sptr_frame), np.min(lerr_sptr_frame), np.mean(lerr_sptr_frame),
                          np.median(lerr_sptr_frame)],
                         [np.max(lerr_ntr_frame), np.min(lerr_ntr_frame), np.mean(lerr_ntr_frame),
                          np.median(lerr_ntr_frame)]]
                mtab4 = pd.DataFrame(metr4, columns=efs_table4_col, index=efs_table4_row)

                # Append extraction metrics to metrics file:
                with open(metric_fname, 'a') as f:
                    print('', file=f)
                    print('Speech enhancement metrics of extraction method using truth:', file=f)
                    print(mtab3.round(2).to_string(), file=f)
                    print('', file=f)
                    print('Truth extraction statistics over frames:', file=f)
                    print(mtab4.round(2).to_string(), file=f)

                # Append to all mixture table
                mtab4flat_col = ['MXSNR', 'SPLERR Max', 'SPLERR Min', 'SPLERR Avg', 'SPLERR Median',
                                 'NLERR Max', 'NLERR Min', 'NLERR Avg', 'NLERR Median']
                # Insert MXSNR
                tmp = np.insert(np.array(metr4), 0, mixdb.mixtures[mi].snr).reshape(1, 9)
                mtab4_flat = pd.DataFrame(tmp, columns=mtab4flat_col, index=[mi])
                if 'all_mtab3' in locals():
                    # Update all mixture table
                    all_mtab3 = pd.concat([all_mtab3, mtab3])
                    all_mtab4 = pd.concat([all_mtab4, mtab4_flat])
                else:
                    # create the table if it doesn't exist
                    all_mtab3 = mtab3
                    all_mtab4 = mtab4_flat

        # 7) write wav files
        if disable_wav is False:
            write_wav(name=basepath + mixidstr + '-mixture.wav', audio=float_to_int16(mixture))
            write_wav(name=basepath + mixidstr + '-speech_est.wav', audio=float_to_int16(speech_est_wav))
            if noise_est_f is not None:
                write_wav(name=basepath + mixidstr + '-noise_est.wav', audio=float_to_int16(noise_est_wav))
            if truth_f is not None:
                write_wav(name=basepath + mixidstr + '-target.wav', audio=float_to_int16(target))
                write_wav(name=basepath + mixidstr + '-noise.wav', audio=float_to_int16(noise))
                if truth_estimate_mode is True:
                    write_wav(name=basepath + mixidstr + '-speech_trest.wav', audio=float_to_int16(speech_trest_wav))
                    write_wav(name=basepath + mixidstr + '-noise_trest.wav', audio=float_to_int16(noise_trest_wav))

        # 8) Write out plot file
        if plot_en is True:
            plot_fname = basepath + mixidstr + '-spenh-plots.pdf'

            # Original size [frames, stride, num_bands]
            # Decimate in the stride dimension
            # Reshape to get frames*decimated_stride, num_bands
            if feature is not None:
                if feature.ndim != 3:
                    raise SonusAIError(f'feature does not have 3 dimensions: frames, stride, num_bands')
                spectrogram = feature[:, -step:, :]
                spectrogram = np.reshape(spectrogram,
                                         (spectrogram.shape[0] * spectrogram.shape[1], spectrogram.shape[2]))
            else:
                spectrogram = None

            with PdfPages(pdf_name) as pdf:
                if mixture is None:
                    pdf.savefig(spec3_plot(predict=10 * np.log10(predict + np.finfo(np.float32).eps),
                                           truth_f=10 * np.log10(truth_f + np.finfo(np.float32).eps),
                                           feature=spectrogram,
                                           title=tfunc_name))
                else:
                    pdf.savefig(spec_energy_plot(mixture=mixture,
                                                 target=target,
                                                 feature=spectrogram,
                                                 truth_f=10 * np.log10(truth_f + np.finfo(np.float32).eps),
                                                 predict=10 * np.log10(predict + np.finfo(np.float32).eps),
                                                 tp_title=tfunc_name))

                    if target_tf is not None:
                        tg_spec = 10 * np.log10(abs2(target_tf) + np.finfo(np.float32).eps)
                        tg_est_spec = 10 * np.log10(abs2(target_est) + np.finfo(np.float32).eps)
                        # n_spec = np.reshape(n_spec,(n_spec.shape[0] * n_spec.shape[1], n_spec.shape[2]))
                        # TODO: change to tg_est when avail
                        pdf.savefig(spec3_plot(predict=tg_est_spec,
                                               truth_f=tg_spec,
                                               feature=spectrogram,
                                               title='target energy_f'))

                    if snr_sd is not None:
                        n_spec = 10 * np.log10(abs2(noise_tf) + np.finfo(np.float32).eps)
                        n_est_spec = 10 * np.log10(abs2(noise_est) + np.finfo(np.float32).eps)
                        pdf.savefig(spec3_plot(predict=n_est_spec,
                                               truth_f=n_spec,
                                               feature=spectrogram,
                                               title='noise energy_f'))

                        pdf.savefig(mix3_plot(mixture=noise_wav_est * 32768,
                                              target=noise,
                                              wav1=snr_sd_frame,
                                              wav2=None,
                                              title='Noise est'))

                        pdf.savefig(mix3_plot(mixture=target_wav_est * 32768,
                                              target=target,
                                              wav1=snr_sd_frame,
                                              wav2=None,
                                              title='Target est'))

            logger.info(f'Wrote plots to {pdf_name}')

    # 9) Done with mixtures, write out summary metrics
    # Calculate SNR summary for
    all_mtab1_sorted = all_mtab1.sort_values(by=['MXSNR', 'SPFILE'])
    all_mtab2_sorted = all_mtab2.sort_values(by=['MXSNR'])
    mtab_snr_summary = all_mtab1_sorted.query('MXSNR==' + str(mixdb.snrs[0])).mean(numeric_only=True).to_frame().T
    mtab_snr_summary_em = all_mtab2_sorted.query('MXSNR==' + str(mixdb.snrs[0])).mean(numeric_only=True).to_frame().T
    for snri in range(1, len(mixdb.snrs)):
        mtab_snr_summary = pd.concat([mtab_snr_summary,
                                      all_mtab1_sorted.query('MXSNR==' + str(mixdb.snrs[snri])).mean(
                                          numeric_only=True).to_frame().T])
        mtab_snr_summary_em = pd.concat([mtab_snr_summary_em,
                                         all_mtab2_sorted.query('MXSNR==' + str(mixdb.snrs[snri])).mean(
                                             numeric_only=True).to_frame().T])

    if truth_estimate_mode is True:
        all_mtab3_sorted = all_mtab3.sort_values(by=['MXSNR'])
        all_mtab4_sorted = all_mtab4.sort_values(by=['MXSNR'])
        mtab_snr_summary_tr = all_mtab1_sorted.query('MXSNR==' + str(mixdb.snrs[0])).mean(
            numeric_only=True).to_frame().T
        mtab_snr_summary_emtr = all_mtab2_sorted.query('MXSNR==' + str(mixdb.snrs[0])).mean(
            numeric_only=True).to_frame().T
        for snri in range(1, len(mixdb.snrs)):
            mtab_snr_summary_tr = pd.concat([mtab_snr_summary_tr,
                                             all_mtab3_sorted.query('MXSNR==' + str(mixdb.snrs[snri])).mean(
                                                 numeric_only=True).to_frame().T])
            mtab_snr_summary_emtr = pd.concat([mtab_snr_summary_emtr,
                                               all_mtab4_sorted.query('MXSNR==' + str(mixdb.snrs[snri])).mean(
                                                   numeric_only=True).to_frame().T])

    if num_mrecords > 1:
        metric_summary_fname = basepath + 'spenh-metrics-summary.txt'
        with open(metric_summary_fname, 'w') as f:
            print(f'Speech enhancement metrics avg over each SNR:', file=f)
            print(mtab_snr_summary.round(2).to_string(), file=f)
            print('', file=f)
            print(f'Extraction statistics stats avg over each SNR:', file=f)
            print(mtab_snr_summary_em.round(2).to_string(), file=f)
            print('', file=f)

            print(f'Speech enhancement metrics stats over all {num_mrecords} mixtures:', file=f)
            print(all_mtab1.describe().round(2).to_string(), file=f)
            print('', file=f)
            print(f'Extraction statistics stats over all {num_mrecords} mixtures:', file=f)
            print(all_mtab2.describe().round(2).to_string(), file=f)
            print('', file=f)

            if truth_estimate_mode is True:
                print(f'Truth-based speech enhancement metrics avg over each SNR:', file=f)
                print(mtab_snr_summary_tr.round(2).to_string(), file=f)
                print('', file=f)
                print(f'Truth-based extraction statistics stats avg over each SNR:', file=f)
                print(mtab_snr_summary_emtr.round(2).to_string(), file=f)
                print('', file=f)

                print(f'Truth-based speech enhancement metrics stats over all {num_mrecords} mixtures:', file=f)
                print(all_mtab3.describe().round(2).to_string(), file=f)
                print('', file=f)
                print(f'Truth-based extraction statistic stats over all {num_mrecords} mixtures:', file=f)
                print(all_mtab4.describe().round(2).to_string(), file=f)
                print('', file=f)

            print('Speech enhancement metrics all-mixtures list:', file=f)
            print(all_mtab1.round(2).to_string(), file=f)
            print('', file=f)
            print('Extraction statistics all-mixtures list:', file=f)
            print(all_mtab2.round(2).to_string(), file=f)

            # Write summary to .csv file
            metric_summary_csvfname = basepath + 'spenh-metrics-summary.csv'
            pd.DataFrame([f'Speech enhancement metrics stats over {num_mrecords} mixtures:']).to_csv \
                (metric_summary_csvfname, mode='a', encoding='utf-8', index=False, header=False)
            all_mtab1.describe().round(2).to_csv(metric_summary_csvfname, encoding='utf-8')
            pd.DataFrame(['']).to_csv(metric_summary_csvfname, mode='a', encoding='utf-8', index=False, header=False)

            pd.DataFrame([f'Extraction statistics stats over {num_mrecords} mixtures:']).to_csv \
                (metric_summary_csvfname, mode='a', encoding='utf-8', index=False, header=False)
            all_mtab2.describe().round(2).to_csv(metric_summary_csvfname, mode='a', encoding='utf-8')
            pd.DataFrame(['']).to_csv(metric_summary_csvfname, mode='a', encoding='utf-8', index=False, header=False)

            pd.DataFrame(['Speech enhancement metrics of extraction method using truth, stats:']).to_csv \
                (metric_summary_csvfname, mode='a', encoding='utf-8', index=False, header=False)
            all_mtab3.describe().round(2).to_csv(metric_summary_csvfname, mode='a', encoding='utf-8')
            pd.DataFrame(['']).to_csv(metric_summary_csvfname, mode='a', encoding='utf-8', index=False, header=False)

            pd.DataFrame(['Truth extraction statistics stats:']).to_csv \
                (metric_summary_csvfname, mode='a', encoding='utf-8', index=False, header=False)
            all_mtab4.describe().round(2).to_csv(metric_summary_csvfname, mode='a', encoding='utf-8')
            pd.DataFrame(['']).to_csv(metric_summary_csvfname, mode='a', encoding='utf-8', index=False, header=False)

            metric_spenhcsv_fname = basepath + 'spenh-metrics-lst.csv'
            pd.DataFrame(['Speech enhancement metrics list:']).to_csv \
                (metric_spenhcsv_fname, mode='a', encoding='utf-8', index=False, header=False)
            all_mtab1.round(2).to_csv(metric_spenhcsv_fname, mode='a', encoding='utf-8')

            metric_estatcsv_fname = basepath + 'spenh-metrics-estats-lst.csv'
            pd.DataFrame(['Extraction statistics list:']).to_csv \
                (metric_estatcsv_fname, mode='a', encoding='utf-8', index=False, header=False)
            all_mtab2.round(2).to_csv(metric_estatcsv_fname, mode='a', encoding='utf-8')

            if truth_estimate_mode is True:
                metric_spenhcsv_fname = basepath + 'spenh-metrics-tru-lst.csv'
                pd.DataFrame(['Speech enhancement metrics list:']).to_csv \
                    (metric_spenhcsv_fname, mode='a', encoding='utf-8', index=False, header=False)
                all_mtab3.round(2).to_csv(metric_spenhcsv_fname, mode='a', encoding='utf-8')

                metric_estatcsv_fname = basepath + 'spenh-metrics-estats-tru-lst.csv'
                pd.DataFrame(['Extraction statistics list:']).to_csv \
                    (metric_estatcsv_fname, mode='a', encoding='utf-8', index=False, header=False)
                all_mtab4.round(2).to_csv(metric_estatcsv_fname, mode='a', encoding='utf-8')

    # if output_name:
    #     with h5py.File(name=output_name, mode='w') as f:
    #         f.create_dataset(name='predict', data=predict)
    #         logger.info(f'Wrote predict data to {output_name}')


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



from typing import Dict
from typing import List
from typing import Union

from sonusai.mixture.types import Location
from sonusai.mixture.types import MRecord
from sonusai.mixture.types import MRecords
from sonusai.mixture.types import MixtureDatabase
from sonusai.mixture.types import MixtureIDs


def load_mixdb(location: Location) -> MixtureDatabase:
    import json
    from os.path import exists
    from os.path import isdir
    from os.path import join

    from sonusai import SonusAIError
    from sonusai.mixture.types import MixtureDatabase

    if not isdir(location):
        raise SonusAIError(f'{location} is not a directory')

    filename = join(location, 'mixdb.json')
    if not exists(filename):
        raise SonusAIError(f'Could not find mixture database in {location}')

    with open(file=filename, mode='r', encoding='utf-8') as f:
        return MixtureDatabase.from_dict(json.loads(f.read()))


def get_samples_in_mixid(mixdb: MixtureDatabase, mixid: int) -> int:
    return mixdb.mixtures[mixid].samples


def get_transform_frames_in_mrecord(mixdb: MixtureDatabase, mrecord: MRecord) -> int:
    from pyaaware import FeatureGenerator

    fg = FeatureGenerator(feature_mode=mixdb.feature)
    return mrecord.samples // fg.ftransform_R


def get_feature_frames_in_mrecord(mixdb: MixtureDatabase, mrecord: MRecord) -> int:
    return mrecord.samples // mixdb.feature_step_samples


def get_class_weights_threshold(mixdb: Union[MixtureDatabase, Dict]) -> List[float]:
    """Get the class_weights_threshold from a mixture database or a config."""
    from sonusai import SonusAIError

    if isinstance(mixdb, dict):
        class_weights_threshold = mixdb['class_weights_threshold']
        num_classes = mixdb['num_classes']
    else:
        class_weights_threshold = mixdb.class_weights_threshold
        num_classes = mixdb.num_classes

    if not isinstance(class_weights_threshold, list):
        class_weights_threshold = [class_weights_threshold] * num_classes

    if len(class_weights_threshold) != num_classes:
        raise SonusAIError(f'invalid class_weights_threshold length: {len(class_weights_threshold)}')

    return class_weights_threshold


def new_mixdb_from_mixids(mixdb: MixtureDatabase, mixids: MixtureIDs) -> MixtureDatabase:
    from copy import deepcopy

    from sonusai import SonusAIError

    mixdb_out = deepcopy(mixdb)
    mixdb_out.mixtures = get_mrecords_from_mixids(mixdb_out, mixids)

    if not mixdb_out.mixtures:
        raise SonusAIError(f'Error processing mixid: {mixids}; resulted in empty list of mixtures')

    return mixdb_out


def convert_mixids_to_list(mixdb: MixtureDatabase, mixids: MixtureIDs = None) -> List[int]:
    from sonusai import SonusAIError

    result = mixids
    all_mixids = list(range(len(mixdb.mixtures)))

    if result is None:
        result = all_mixids

    if isinstance(result, str):
        if result == '*':
            result = all_mixids
        else:
            try:
                result = eval(f'{all_mixids}[{result}]')
            except NameError:
                result = []

    if isinstance(result, range):
        result = list(result)

    if isinstance(result, int):
        result = [result]

    if not all(isinstance(x, int) and 0 <= x < len(mixdb.mixtures) for x in result):
        raise SonusAIError(f'Invalid entries in mixids of {mixids}')

    if not result:
        raise SonusAIError(f'Empty mixids {mixids}')

    return result


def get_mrecords_from_mixids(mixdb: MixtureDatabase, mixids: MixtureIDs = None) -> MRecords:
    from copy import deepcopy

    return [deepcopy(mixdb.mixtures[i]) for i in convert_mixids_to_list(mixdb, mixids)]


def evaluate_random_rule(rule: str) -> Union[str, float]:
    """Evaluate 'rand' directive."""
    import re
    from random import uniform

    from sonusai.mixture.constants import RAND_PATTERN

    def rand_repl(m):
        return f'{uniform(float(m.group(1)), float(m.group(4))):.2f}'

    return eval(re.sub(RAND_PATTERN, rand_repl, rule))


def generate_mixture_filename(mixid: int, padding: int) -> str:
    return f'{mixid:0{padding}}.h5'


def get_mixture_metadata(mixdb: MixtureDatabase, mixid: int) -> str:
    mrecord = mixdb.mixtures[mixid]
    metadata = ''
    for ti in range(len(mrecord.target_file_index)):
        tfi = mrecord.target_file_index[ti]
        tai = mrecord.target_augmentation_index[ti]
        metadata += f'target {ti} name: {mixdb.targets[tfi].name}\n'
        metadata += f'target {ti} augmentation: {mixdb.target_augmentations[tai].to_dict()}\n'
        metadata += f'target {ti} target_gain: {mrecord.target_gain[ti]}\n'
        truth_settings = mixdb.targets[tfi].truth_settings
        for tsi in range(len(truth_settings)):
            metadata += f'target {ti} truth index {tsi}: {truth_settings[tsi].index}\n'
            metadata += f'target {ti} truth function {tsi}: {truth_settings[tsi].function}\n'
            metadata += f'target {ti} truth config {tsi}: {truth_settings[tsi].config}\n'
    metadata += f'noise name: {mixdb.noises[mrecord.noise_file_index].name}\n'
    metadata += f'noise augmentation: {mixdb.noise_augmentations[mrecord.noise_augmentation_index].to_dict()}\n'
    metadata += f'snr: {mrecord.snr}\n'
    metadata += f'random_snr: {mrecord.random_snr}\n'
    metadata += f'samples: {mrecord.samples}\n'
    metadata += f'target_snr_gain: {mrecord.target_snr_gain}\n'
    metadata += f'noise_snr_gain: {mrecord.noise_snr_gain}\n'

    return metadata



# SonusAI mixture utilities
from sonusai.mixture.audio import add_audio_data_to_h5
from sonusai.mixture.audio import calculate_audio_from_transform
from sonusai.mixture.audio import calculate_transform_from_audio
from sonusai.mixture.audio import check_audio_files_exist
from sonusai.mixture.audio import concatenate_audio_data
from sonusai.mixture.audio import get_audio_data
from sonusai.mixture.audio import get_augmented_noise_audios
from sonusai.mixture.audio import get_mixture_data_deprecated
from sonusai.mixture.audio import get_next_noise
from sonusai.mixture.audio import get_raw_audios
from sonusai.mixture.audio import get_target_noise_audio
from sonusai.mixture.audio import read_audio
from sonusai.mixture.augmentation import apply_augmentation
from sonusai.mixture.augmentation import apply_snr_gain
from sonusai.mixture.augmentation import estimate_audio_length
from sonusai.mixture.augmentation import get_augmentation_indices_for_mixup
from sonusai.mixture.augmentation import get_augmentations
from sonusai.mixture.augmentation import get_mixups
from sonusai.mixture.augmentation import get_pad_length
from sonusai.mixture.augmentation import pad_to_samples
from sonusai.mixture.balance import get_class_balancing_augmentation
from sonusai.mixture.class_count import get_class_count_from_mixids
from sonusai.mixture.class_count import get_class_count_from_mixtures
from sonusai.mixture.class_count import get_class_count_from_mrecord
from sonusai.mixture.config import get_default_config
from sonusai.mixture.config import get_max_class
from sonusai.mixture.config import get_noise_files
from sonusai.mixture.config import get_target_files
from sonusai.mixture.config import load_config
from sonusai.mixture.config import raw_load_config
from sonusai.mixture.config import tokenized_expandvars
from sonusai.mixture.config import update_config_from_file
from sonusai.mixture.config import update_config_from_hierarchy
from sonusai.mixture.config import update_truth_settings
from sonusai.mixture.constants import BIT_DEPTH
from sonusai.mixture.constants import CHANNEL_COUNT
from sonusai.mixture.constants import DEFAULT_CONFIG
from sonusai.mixture.constants import DEFAULT_NOISE
from sonusai.mixture.constants import FLOAT_BYTES
from sonusai.mixture.constants import RAND_PATTERN
from sonusai.mixture.constants import REQUIRED_CONFIGS
from sonusai.mixture.constants import SAMPLE_BYTES
from sonusai.mixture.constants import SAMPLE_RATE
from sonusai.mixture.constants import VALID_AUGMENTATIONS
from sonusai.mixture.constants import VALID_CONFIGS
from sonusai.mixture.constants import VALID_TRUTH_SETTINGS
from sonusai.mixture.feature import add_feature_data_to_h5
from sonusai.mixture.feature import concatenate_feature_data
from sonusai.mixture.feature import get_feature_data
from sonusai.mixture.feature import get_feature_data_from_dir
from sonusai.mixture.feature import get_feature_from_audio
from sonusai.mixture.feature import get_feature_stats
from sonusai.mixture.log_duration_and_sizes import log_duration_and_sizes
from sonusai.mixture.mapped_snr_f import calculate_mapped_snr_f
from sonusai.mixture.mapped_snr_f import calculate_snr_f_statistics
from sonusai.mixture.mixdb import convert_mixids_to_list
from sonusai.mixture.mixdb import evaluate_random_rule
from sonusai.mixture.mixdb import generate_mixture_filename
from sonusai.mixture.mixdb import get_class_weights_threshold
from sonusai.mixture.mixdb import get_feature_frames_in_mrecord
from sonusai.mixture.mixdb import get_mixture_metadata
from sonusai.mixture.mixdb import get_mrecords_from_mixids
from sonusai.mixture.mixdb import get_samples_in_mixid
from sonusai.mixture.mixdb import get_transform_frames_in_mrecord
from sonusai.mixture.mixdb import load_mixdb
from sonusai.mixture.mixdb import new_mixdb_from_mixids
from sonusai.mixture.process import process_mixture
from sonusai.mixture.process import process_target
from sonusai.mixture.segsnr import generate_segsnr
from sonusai.mixture.segsnr import segsnr_sample_to_frame
from sonusai.mixture.target_class_balancing import balance_targets
from sonusai.mixture.targets import get_augmented_target_indices_by_class
from sonusai.mixture.targets import get_augmented_target_indices_for_mixup
from sonusai.mixture.targets import get_augmented_targets
from sonusai.mixture.targets import get_target_indices_for_truth_index
from sonusai.mixture.targets import get_truth_indices_for_target
from sonusai.mixture.truth import generate_truth
from sonusai.mixture.truth import get_truth_indices_for_mixid
from sonusai.mixture.truth import truth_function
from sonusai.mixture.truth import truth_reduction
from sonusai.mixture.types import AudioData
from sonusai.mixture.types import AudioF
from sonusai.mixture.types import AudioT
from sonusai.mixture.types import AudiosF
from sonusai.mixture.types import AudiosT
from sonusai.mixture.types import Augmentation
from sonusai.mixture.types import Augmentations
from sonusai.mixture.types import AugmentedTarget
from sonusai.mixture.types import AugmentedTargets
from sonusai.mixture.types import Feature
from sonusai.mixture.types import FeatureData
from sonusai.mixture.types import ListAudiosT
from sonusai.mixture.types import Location
from sonusai.mixture.types import MRecord
from sonusai.mixture.types import MRecords
from sonusai.mixture.types import MixtureDatabase
from sonusai.mixture.types import MixtureIDs
from sonusai.mixture.types import Predict
from sonusai.mixture.types import Segsnr
from sonusai.mixture.types import TargetFile
from sonusai.mixture.types import TargetFiles
from sonusai.mixture.types import Truth
from sonusai.mixture.types import TruthFunctionConfig
from sonusai.mixture.types import TruthSettings



import numpy as np

from sonusai.mixture.types import AudioT
from sonusai.mixture.types import MRecord
from sonusai.mixture.types import MixtureDatabase
from sonusai.mixture.types import Segsnr


def generate_segsnr(mixdb: MixtureDatabase,
                    mrecord: MRecord,
                    target_audio: AudioT,
                    noise_audio: AudioT,
                    compute: bool = True,
                    frame_based: bool = False) -> Segsnr:
    """Generate segmental SNR."""
    from pyaaware import FeatureGenerator
    from pyaaware import ForwardTransform

    from sonusai.utils import int16_to_float

    if not compute:
        return np.empty(0, dtype=np.float32)

    fg = FeatureGenerator(feature_mode=mixdb.feature)
    fft = ForwardTransform(N=fg.ftransform_N, R=fg.ftransform_R, ttype=fg.ftransform_ttype)

    if frame_based:
        segsnr = np.empty(mrecord.samples // fg.ftransform_R, dtype=np.float32)
    else:
        segsnr = np.empty(mrecord.samples, dtype=np.float32)

    frame = 0
    for offset in range(0, mrecord.samples, fg.ftransform_R):
        indices = slice(offset, offset + fg.ftransform_R)

        target_energy = fft.energy_t(int16_to_float(target_audio[indices]))
        noise_energy = fft.energy_t(int16_to_float(noise_audio[indices]))

        if noise_energy == 0:
            snr = np.float32(np.inf)
        else:
            snr = np.float32(target_energy / noise_energy)

        if frame_based:
            segsnr[frame] = snr
            frame += 1
        else:
            segsnr[indices] = snr

    return segsnr


def segsnr_sample_to_frame(mixdb: MixtureDatabase, segsnr_t: Segsnr) -> Segsnr:
    from pyaaware import FeatureGenerator

    fg = FeatureGenerator(feature_mode=mixdb.feature)

    return segsnr_t[0::fg.ftransform_R]



from typing import List

import numpy as np

from sonusai.mixture.types import AudioT
from sonusai.mixture.types import AudiosT
from sonusai.mixture.types import MRecord
from sonusai.mixture.types import MixtureDatabase
from sonusai.mixture.types import Truth
from sonusai.mixture.types import TruthFunctionConfig


def _strictly_decreasing(list_to_check: list) -> bool:
    return all(x > y for x, y in zip(list_to_check, list_to_check[1:]))


def generate_truth(mixdb: MixtureDatabase,
                   mrecord: MRecord,
                   target_audios: AudiosT,
                   noise_audio: AudioT,
                   compute: bool = True) -> Truth:
    from sonusai import SonusAIError

    if not compute:
        return np.empty(0, dtype=np.float32)

    if len(target_audios) != len(mrecord.target_file_index):
        raise SonusAIError('Number of target audio entries does not match number of targets')

    if not all(len(target_audio) == len(noise_audio) for target_audio in target_audios):
        raise SonusAIError('Lengths of target audio do not match length of noise audio')

    truth = np.zeros((mrecord.samples, mixdb.num_classes), dtype=np.float32)
    for idx in range(len(target_audios)):
        for truth_setting in mixdb.targets[mrecord.target_file_index[idx]].truth_settings:
            config = TruthFunctionConfig(
                feature=mixdb.feature,
                index=truth_setting.index,
                function=truth_setting.function,
                config=truth_setting.config,
                num_classes=mixdb.num_classes,
                mutex=mixdb.truth_mutex,
                target_gain=mrecord.target_gain[idx] * mrecord.target_snr_gain
            )
            truth += truth_function(target_audio=target_audios[idx], noise_audio=noise_audio, config=config)

    return truth


def truth_function(target_audio: AudioT, noise_audio: AudioT, config: TruthFunctionConfig) -> Truth:
    import h5py
    from pyaaware import FeatureGenerator
    from pyaaware import ForwardTransform
    from pyaaware import SED

    from sonusai import SonusAIError
    from sonusai.mixture.mapped_snr_f import calculate_mapped_snr_f
    from sonusai.utils.numeric_conversion import int16_to_float

    fg = FeatureGenerator(feature_mode=config.feature, num_classes=config.num_classes, truth_mutex=config.mutex)
    frame_size = fg.ftransform_R
    truth = np.zeros((len(target_audio), config.num_classes), dtype=np.float32)
    offsets = range(0, len(target_audio), frame_size)
    zero_based_indices = [x - 1 for x in config.index]
    target_fft = ForwardTransform(N=fg.ftransform_N, R=fg.ftransform_R, ttype=fg.ftransform_ttype)
    noise_fft = ForwardTransform(N=fg.ftransform_N, R=fg.ftransform_R, ttype=fg.ftransform_ttype)

    if config.function == 'sed':
        if len(target_audio) % frame_size != 0:
            raise SonusAIError(f'Number of samples in audio is not a multiple of {frame_size}')

        if config.config is None:
            raise SonusAIError('Truth function SED missing config')

        parameters = ['thresholds']
        for parameter in parameters:
            if parameter not in config.config:
                raise SonusAIError(f'Truth function SED config missing required parameter: {parameter}')

        thresholds = config.config['thresholds']
        if not _strictly_decreasing(thresholds):
            raise SonusAIError(f'Truth function SED thresholds are not strictly decreasing: {thresholds}')

        if config.target_gain == 0:
            return truth

        # SED wants 1-based indices
        sed = SED(thresholds=thresholds,
                  index=config.index,
                  frame_size=frame_size,
                  num_classes=config.num_classes,
                  mutex=config.mutex)

        target_audio = np.int16(np.round(np.float32(target_audio) / config.target_gain))
        for offset in offsets:
            indices = slice(offset, offset + frame_size)
            new_truth = sed.execute(target_fft.energy_t(int16_to_float(target_audio[indices])))
            truth[indices] = np.reshape(new_truth, (1, len(new_truth)))

        return truth

    if config.function == 'file':
        if config.config is None:
            raise SonusAIError('Truth function file missing config')

        parameters = ['file']
        for parameter in parameters:
            if parameter not in config.config:
                raise SonusAIError(f'Truth function file config missing required parameter: {parameter}')

        with h5py.File(config.config['file'], 'r') as f:
            truth_in = np.array(f['truth_t'])

        if truth_in.ndim != 2:
            raise SonusAIError('Truth file data is not 2 dimensions')

        if truth_in.shape[0] != len(target_audio):
            raise SonusAIError('Truth file does not contain the right amount of samples')

        if config.target_gain == 0:
            return truth

        if len(zero_based_indices) > 1:
            if len(zero_based_indices) != truth_in.shape[1]:
                raise SonusAIError('Truth file does not contain the right amount of classes')

            truth[:, zero_based_indices] = truth_in
        else:
            index = zero_based_indices[0]
            if index + truth_in.shape[1] > config.num_classes:
                raise SonusAIError('Truth file contains too many classes')

            truth[:, index:index + truth_in.shape[1]] = truth_in

        return truth

    if config.function == 'energy_t':
        if config.target_gain == 0:
            return truth

        for offset in offsets:
            target_energy = target_fft.energy_t(int16_to_float(target_audio[offset:offset + frame_size]))
            truth[offset:offset + frame_size, zero_based_indices] = np.float32(target_energy)

        return truth

    if config.function == 'target_f':
        truth = np.zeros((len(target_audio), config.num_classes), dtype=np.float32)
        if config.target_gain == 0:
            return truth

        for offset in offsets:
            target_f = np.complex64(
                target_fft.execute(int16_to_float(target_audio[offset:offset + frame_size])))

            indices = slice(offset, offset + frame_size)
            for index in zero_based_indices:
                truth[indices, index:index + target_fft.bins] = np.real(target_f)
                truth[indices, (index + target_fft.bins):(index + 2 * target_fft.bins)] = np.imag(target_f)

        return truth

    if config.function == 'crm':
        truth = np.zeros((len(target_audio), config.num_classes), dtype=np.float32)

        if config.target_gain == 0:
            return truth

        for offset in offsets:
            target_f = np.complex64(
                target_fft.execute(int16_to_float(target_audio[offset:offset + frame_size])))
            mixture_f = np.complex64(
                noise_fft.execute(int16_to_float(noise_audio[offset:offset + frame_size]))) + target_f

            crm = np.empty(target_f.shape, dtype=np.complex64)
            with np.nditer(target_f, flags=['multi_index'], op_flags=['readwrite']) as it:
                for _ in it:
                    num = target_f[it.multi_index]
                    den = mixture_f[it.multi_index]
                    if num == 0:
                        crm[it.multi_index] = 0
                    elif den == 0:
                        crm[it.multi_index] = complex(np.inf, np.inf)
                    else:
                        crm[it.multi_index] = num / den

            indices = slice(offset, offset + frame_size)
            for index in zero_based_indices:
                truth[indices, index:index + target_fft.bins] = np.real(crm)
                truth[indices, (index + target_fft.bins):(index + 2 * target_fft.bins)] = np.imag(crm)

        return truth

    if config.function in ['energy_f', 'snr_f', 'mapped_snr_f']:
        if config.target_gain == 0:
            return truth

        snr_db_mean = None
        snr_db_std = None
        if config.function == 'mapped_snr_f':
            if config.config is None:
                raise SonusAIError('Truth function mapped SNR missing config')

            parameters = ['snr_db_mean', 'snr_db_std']
            for parameter in parameters:
                if parameter not in config.config:
                    raise SonusAIError(f'Truth function mapped SNR config missing required parameter: {parameter}')

            snr_db_mean = config.config['snr_db_mean']
            if len(snr_db_mean) != target_fft.bins:
                raise SonusAIError(f'Truth function mapped SNR snr_db_mean does not have {target_fft.bins} elements')

            snr_db_std = config.config['snr_db_std']
            if len(snr_db_std) != target_fft.bins:
                raise SonusAIError(f'Truth function mapped SNR snr_db_std does not have {target_fft.bins} elements')

        for index in zero_based_indices:
            if index + target_fft.bins > config.num_classes:
                raise SonusAIError('Truth index exceeds the number of classes')

        for offset in offsets:
            target_energy = np.float32(
                target_fft.energy_f(int16_to_float(target_audio[offset:offset + frame_size])))
            if config.function in ['snr_f', 'mapped_snr_f']:
                noise_energy = np.float32(
                    noise_fft.energy_f(int16_to_float(noise_audio[offset:offset + frame_size])))
            else:
                noise_energy = np.float32(1)

            indices = slice(offset, offset + frame_size)
            for index in zero_based_indices:
                old_err = np.seterr(divide='ignore', invalid='ignore')
                tmp = target_energy / noise_energy
                np.seterr(**old_err)

                tmp = np.nan_to_num(tmp, nan=-np.inf, posinf=np.inf, neginf=-np.inf)

                if config.function == 'mapped_snr_f':
                    tmp = calculate_mapped_snr_f(tmp, snr_db_mean, snr_db_std)

                truth[indices, index:index + target_fft.bins] = tmp

        return truth

    if config.function == 'phoneme':
        # Read in .txt transcript and run a Python function to generate text grid data
        # (indicating which phonemes are active)
        # Then generate truth based on this data and put in the correct classes based on config.index
        raise SonusAIError('Truth function phoneme is not supported yet')

    raise SonusAIError(f'Unsupported truth function: {config.function}')


def get_truth_indices_for_mixid(mixdb: MixtureDatabase, mixid: int) -> List[int]:
    """Get a list of truth indices for a given mixid."""

    from sonusai.mixture.targets import get_truth_indices_for_target

    indices = []
    for target_file_index in mixdb.mixtures[mixid].target_file_index:
        indices.append(*get_truth_indices_for_target(mixdb.targets[target_file_index]))

    return sorted(list(set(indices)))


def truth_reduction(x: Truth, func: str) -> Truth:
    from sonusai import SonusAIError

    if func == 'max':
        return np.max(x, axis=0)

    if func == 'mean':
        return np.mean(x, axis=0)

    raise SonusAIError(f'Invalid truth reduction function: {func}')
